"""Code is adapted from https://github.com/earthnet2021/earthnet-toolkit."""
from typing import Tuple, Optional, Sequence, Union
import argparse
import re
import json
import time
import numpy as np
from skimage import metrics
import scipy.stats
from pathlib import Path
import multiprocessing
from tqdm import tqdm
import warnings


class CubeCalculator:
    """Loads single cube and calculates subscores for EarthNetScore

    Example:

        >>> scores = CubeCalculator.get_scores({"pred_filepath": Path/to/pred.npz, "targ_filepath": Path/to/targ.npz})
    """    

    @staticmethod
    def MAD(preds: np.ndarray, targs: np.ndarray, masks: np.ndarray) -> Tuple[float, dict]:
        """Median absolute deviation score

        Median absolute deviation between non-masked target and predicted pixels. Scaled by a scaling factor such that a distance the size of a 99.7% confidence interval of the variance of the pixelwise centered timeseries is scaled to 0.9 (such that the mad-score becomes 0.1). The mad-score is 1-MAD, it is scaled from 0 (worst) to 1 (best).

        Args:
            preds (np.ndarray): Predictions, shape h,w,c,t
            targs (np.ndarray): Targets, shape h,w,c,t
            masks (np.ndarray): Masks, shape h,w,c,t, 1 if non-masked, else 0
        Returns:
            Tuple[float, dict]: mad-score, debugging information
        """        
        dists = np.abs(preds-targs)
        dists[masks == 0] = np.nan

        scaling_factor = 0.06649346971087526 # Computed via the expected distance from pixelwise timeseries variance
        dists = dists.astype(np.float64)

        scaled_dists = dists ** scaling_factor

        distmedian = np.nanmedian(scaled_dists)
        if distmedian is None:
            mad = None
        else:
            mad = max(0,min(1,1-distmedian))
        
        MAE_frames = []
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            for t in range(dists.shape[-1]):
                mean = np.nanmean(dists[:,:,:,t])
                if mean is np.nan:
                    mean = 1000
                MAE_frames.append(mean)
        
        debug_info = {
                        "minimum distance": float(np.nanmin(dists)), 
                        "maximum distance": float(np.nanmax(dists)), 
                        "mean distance": float(np.nanmean(dists)), 
                        "median distance": float(np.nanmedian(dists)), 
                        "number nan": float(np.isnan(dists).sum()), 
                        "MAD score": float(mad),
                        "frames": MAE_frames 
                    }
        return mad, debug_info

    @staticmethod
    def OLS(preds: np.ndarray, targs: np.ndarray, masks: np.ndarray) -> Tuple[float, dict]:
        """Ordinary least squares slope deviation score

        Mean absolute difference between ordinary least squares slopes of target and predicted pixelwise NDVI timeseries. Target slopes are calculated over non-masked values. Predicted slopes are calculated for all values between the first and last non-masked value of a given timeseries. Scaled by a scaling factor such that a distance the size of a 99.7% confidence interval of the variance of the pixelwise centered NDVI timeseries is scaled to 0.9 (such that the ols-score becomes 0.1). If the timeseries is longer than 40 steps, it is split up into parts of length 20. The ols-score is 1-mean(abs(b_targ - b_pred)), it is scaled from 0 (worst) to 1 (best).

        Args:
            preds (np.ndarray): NDVI Predictions, shape h,w,1,t
            targs (np.ndarray): NDVI Targets, shape h,w,1,t
            masks (np.ndarray): NDVI Masks, shape h,w,1,t, 1 if non-masked, else 0

        Returns:
            Tuple[float, dict]: ols-score, debugging information
        """        

        h, w, c, t = preds.shape


        if t > 40: # Checking if pixelwise timeseries is too long
            assert(t%20 == 0)
            preds = np.reshape(preds, (h, w, -1, 20))
            targs = np.reshape(targs, (h, w, -1, 20))
            masks = np.reshape(masks, (h, w, -1, 20))
            h, w, c, t = preds.shape
        
        A = np.vstack([np.linspace(1,2,t),np.ones(t)]).T[np.newaxis,:,:].repeat(c*h*w, 0)
        targs = np.reshape(targs, (-1, t))[:,:,np.newaxis]
        preds = np.reshape(preds, (-1, t))[:,:,np.newaxis]
        masks = np.reshape(masks, (-1, t))[:,:,np.newaxis]
        masks[(masks.sum(1, keepdims = True) < 2).repeat(t,1)] = 0
        targsmasked = targs * masks

        Atarg = A * masks
        Atargmin = np.ma.masked_equal(Atarg, 0.0, copy=True)
        Atarg[:,:,0] = np.where(Atarg[:,:,0] > 0, Atarg[:,:,0] - Atargmin[:,:,0].min(1,keepdims = True), Atarg[:,:,0]) 
        Atarg[:,:,0] = 2 * np.where(Atarg[:,:,1] > 0, Atarg[:,:,0] / (Atarg[:,:,0].max(1, keepdims = True) + 1e-8) + 1, Atarg[:,:,0])
        AtargT = Atarg.transpose(0,2,1)

        predmask = np.where((A[:,:,0] >= Atargmin[:,:,0].min(1,keepdims = True)) & (A[:,:,0] <= Atargmin[:,:,0].max(1,keepdims = True)), np.ones_like(A[:,:,0]), np.zeros_like(A[:,:,0]))[:,:,np.newaxis]
        A = A * predmask
        A[:,:,0] = np.where(A[:,:,0] > 0, A[:,:,0] - Atargmin[:,:,0].min(1,keepdims = True), A[:,:,0]) 
        A[:,:,0] = 2 * np.where(A[:,:,1] > 0, A[:,:,0] / (A[:,:,0].max(1, keepdims = True) + 1e-8) + 1, A[:,:,0])

        predsmasked = preds * predmask

        AT = A.transpose(0,2,1)

        noise = np.random.rand(c*h*w,2,2)/10000

        btarg = np.matmul(np.linalg.inv(np.matmul(AtargT,Atarg)+noise),np.matmul(AtargT,targsmasked))
        
        bpred = np.matmul(np.linalg.inv(np.matmul(AT ,A) + noise),np.matmul(AT,predsmasked))      

        dists = np.abs(btarg[:,0,0] - bpred[:,0,0])/2

        scaling_factor = 0.10082047548620601 # Computed via the expected distance from pixelwise timeseries variance
        
        dists = dists.astype(np.float64)

        scaled_dists = dists ** scaling_factor
        
        distmean = np.nanmean(scaled_dists)
        if distmean is None:
            ols = None
        else:
            ols = max(0,min(1, 1-distmean))

        debug_info = {
                        #"target slopes": btarg[:,0,0].tolist(), 
                        #"predicted slopes": bpred[:,0,0].tolist(), 
                        "min target slope": float(np.nanmin(btarg[:,0,0])), 
                        "max target slope": float(np.nanmax(btarg[:,0,0])), 
                        "mean target slope": float(np.nanmean(btarg[:,0,0])), 
                        "number nan target slope": float(np.isnan(btarg[:,0,0]).sum()),
                        "min pred slope": float(np.nanmin(bpred[:,0,0])), 
                        "max pred slope": float(np.nanmax(bpred[:,0,0])), 
                        "mean pred slope": float(np.nanmean(bpred[:,0,0])), 
                        "number nan pred slope": float(np.isnan(bpred[:,0,0]).sum()),
                        "minimum distance": float(np.nanmin(dists)), 
                        "maximum distance": float(np.nanmax(dists)), 
                        "mean distance": float(np.nanmean(dists)), 
                        "median distance": float(np.nanmedian(dists)), 
                        "number nan": float(np.isnan(dists).sum()), 
                        "ols score": float(ols)}

        return ols, debug_info

    @classmethod
    def EMD(cls, preds: np.ndarray, targs: np.ndarray, masks: np.ndarray) -> Tuple[float, dict]:
        """Earth mover distance score

        The earth mover distance (w1 metric) is computed between target and predicted pixelwise NDVI timeseries value distributions. For the target distributions, only non-masked values are considered. Scaled by a scaling factor such that a distance the size of a 99.7% confidence interval of the variance of the pixelwise centered NDVI timeseries is scaled to 0.9 (such that the ols-score becomes 0.1). The emd-score is 1-mean(emd), it is scaled from 0 (worst) to 1 (best).

        Args:
            preds (np.ndarray): NDVI Predictions, shape h,w,1,t
            targs (np.ndarray): NDVI Targets, shape h,w,1,t
            masks (np.ndarray): NDVI Masks, shape h,w,1,t, 1 if non-masked, else 0

        Returns:
            Tuple[float, dict]: emd-score, debugging information
        """        
        data = np.concatenate([preds, targs, masks], axis = -1)
        dists = np.apply_along_axis(cls.compute_w1, axis = -1, arr = data)
        
        scaling_factor = 0.10082047548620601 # Computed via the expected distance from pixelwise timeseries variance

        dists = np.abs(dists).astype(np.float64)

        scaled_dists = dists ** scaling_factor

        distmean = np.nanmean(scaled_dists)

        if distmean is None:
            emd = None
        else:
            emd = max(0,min(1, 1-distmean))

        debug_info = {
                        "minimum distance": float(np.nanmin(dists)), 
                        "maximum distance": float(np.nanmax(dists)), 
                        "mean distance": float(np.nanmean(dists)), 
                        "median distance": float(np.nanmedian(dists)), 
                        "number nan": float(np.isnan(dists).sum()), 
                        "w1 score": float(emd)
                    }

        return emd, debug_info

    @staticmethod
    def compute_w1(datarow: np.ndarray) -> Union[np.ndarray, None]:
        """Computing w1 distance for np.apply_along_axis

        Args:
            datarow (np.ndarray): 1-dimensional array that can be split into three parts of equal size, these are in order: predictions, targets and masks for a single pixel and channel through time.

        Returns:
            Union[np.ndarray, None]: w1 distance between prediction and target, if not completely masked, else None.
        """        
        preds, targs, masks = np.split(datarow, 3)
        targs = targs[masks == 1]
        if len(targs) > 1:
            return scipy.stats.wasserstein_distance(preds, targs)
        else:
            return np.nan

    @staticmethod
    def SSIM(preds: np.ndarray, targs: np.ndarray, masks: np.ndarray) -> Tuple[float, dict]:
        """Structural similarity index score

        Structural similarity between predicted and target cube computed for all channels and frames individually if the given target is less than 30% masked. Scaled by a scaling factor such that a mean SSIM of 0.8 is scaled to a ssim-score of 0.1. The ssim-score is mean(ssim), it is scaled from 0 (worst) to 1 (best).

        Args:
            preds (np.ndarray): Predictions, shape h,w,c,t
            targs (np.ndarray): Targets, shape h,w,c,t
            masks (np.ndarray): Masks, shape h,w,c,t, 1 if non-masked, else 0

        Returns:
            Tuple[float, dict]: ssim-score, debugging information
        """        

        ssim_targs = np.where(masks, targs, preds)
        new_shape = (-1, preds.shape[0], preds.shape[1])
        ssim_targs = np.transpose(np.reshape(np.transpose(ssim_targs, (3,2,0,1)), new_shape),(1,2,0))
        ssim_preds = np.transpose(np.reshape(np.transpose(preds, (3,2,0,1)), new_shape),(1,2,0))
        ssim_masks = np.transpose(np.reshape(np.transpose(masks, (3,2,0,1)), new_shape),(1,2,0))
        running_ssim = 0
        counts = 0
        ssim_frames = []
        for i in range(ssim_targs.shape[-1]):
            if ssim_masks[:,:,i].sum() > 0.7*ssim_masks[:,:,i].size:
                curr_ssim = metrics.structural_similarity(ssim_targs[:,:,i], ssim_preds[:,:,i])
                running_ssim += curr_ssim
                counts += 1
            else:
                curr_ssim = 1000
            ssim_frames.append(curr_ssim)
        
        if counts == 0:
            ssim = None
        else:
            ssim = max(0,(running_ssim/max(counts,1)))

            scaling_factor = 10.31885115 # Scales SSIM=0.8 down to 0.1

            ssim = float(ssim ** scaling_factor)
        
        debug_info = {
                        #"framewise SSIM, 1000 if frame was too much masked": ssim_frames, 
                        "Min SSIM": str(np.ma.filled(np.ma.masked_equal(np.array(ssim_frames), 1000.0).min(), np.nan)),
                        "Max SSIM": str(np.ma.filled(np.ma.masked_equal(np.array(ssim_frames), 1000.0).max(), np.nan)),
                        "Mean SSIM": str(np.ma.filled(np.ma.masked_equal(np.array(ssim_frames), 1000.0).mean(), np.nan)),
                        "Standard deviation SSIM": str(np.ma.filled(np.ma.masked_equal(np.array(ssim_frames), 1000.0).std(), np.nan)),
                        "Valid SSIM frames": counts,
                        "SSIM score": ssim,
                        "frames": ssim_frames
                    }

        return ssim, debug_info

    @staticmethod
    def load_file(pred_filepath: Path, targ_filepath: Path) -> Sequence[np.ndarray]:
        """Load a single target cube and a matching prediction

        Args:
            pred_filepath (Path): Path to predicted cube
            targ_filepath (Path): Path to target cube

        Returns:
            Sequence[np.ndarray]: preds, targs, masks, ndvi_preds, ndvi_targs, ndvi_masks
        """        
        
        pred_npz = np.load(pred_filepath)
        targ_npz = np.load(targ_filepath)

        pred_key = "highresdynamic" if "highresdynamic" in pred_npz.keys() else list(pred_npz.keys())[0]

        preds = pred_npz[pred_key][:,:,:4,:]
        targs = targ_npz["highresdynamic"][:,:,:4,:]
        masks = ((1 - targ_npz["highresdynamic"][:,:,-1,:])[:,:,np.newaxis,:]).repeat(4,2)

        if preds.shape[-1] < targs.shape[-1]:
            targs = targs[:,:,:,-preds.shape[-1]:]
            masks = masks[:,:,:,-preds.shape[-1]:]
        
        assert(preds.shape == targs.shape)

        preds[preds < 0] = 0
        preds[preds > 1] = 1

        targs[np.isnan(targs)] = 0
        targs[targs > 1] = 1
        targs[targs < 0] = 0

        ndvi_preds = ((preds[:,:,3,:] - preds[:,:,2,:])/(preds[:,:,3,:] + preds[:,:,2,:] + 1e-6))[:,:,np.newaxis,:]
        ndvi_targs = ((targs[:,:,3,:] - targs[:,:,2,:])/(targs[:,:,3,:] + targs[:,:,2,:] + 1e-6))[:,:,np.newaxis,:]
        ndvi_masks = masks[:,:,0,:][:,:,np.newaxis,:]

        return preds, targs, masks, ndvi_preds, ndvi_targs, ndvi_masks

    @classmethod
    def get_scores(cls, filepaths: dict) -> dict:
        """Get all subscores for a given cube

        Args:
            filepaths (dict): Has keys "pred_filepath", "targ_filepath" with respective paths.

        Returns:
            dict: subscores and debugging info for the input cube
        """        
        assert({"pred_filepath", "targ_filepath"}.issubset(set(filepaths.keys())))
        
        preds, targs, masks, ndvi_preds, ndvi_targs, ndvi_masks = cls.load_file(filepaths["pred_filepath"], filepaths["targ_filepath"])

        debug_info = {}

        mad, debug_info["MAD"] = cls.MAD(preds, targs, masks)

        ols, debug_info["OLS"] = cls.OLS(ndvi_preds, ndvi_targs, ndvi_masks)

        emd, debug_info["EMD"] = cls.EMD(ndvi_preds, ndvi_targs, ndvi_masks)

        ssim, debug_info["SSIM"] = cls.SSIM(preds, targs, masks)

        return {
            "pred_filepath": str(filepaths["pred_filepath"]),
            "targ_filepath": str(filepaths["targ_filepath"]),
            "MAD": mad,
            "OLS": ols,
            "EMD": emd,
            "SSIM": ssim,
            "debug_info": debug_info
        }

class EarthNetScore:
    """EarthNetScore class, fast computation using multiprocessing

    Example:

        Direct computation
        >>> EarthNetScore.get_ENS(Path/to/predictions, Path/to/targets, data_output_file = Path/to/data.json, ens_output_file = Path/to/ens.json)

        More control (for further plotting)
        >>> ENS = EarthNetScore(Path/to/predictions, Path/to/targets)
        >>> data = ENS.compute_scores()
        >>> ens = ENS.summarize()

    """    
    def __init__(self, pred_dir: str, targ_dir: str):
        """Initialize EarthNetScore

        Args:
            pred_dir (str): Directory with predictions, format is one of {pred_dir/tile/cubename.npz, pred_dir/tile/experiment_cubename.npz}
            targ_dir (str): Directory with targets, format is one of {targ_dir/target/tile/target_cubename.npz, targ_dir/target/tile/cubename.npz, targ_dir/tile/target_cubename.npz, targ_dir/tile/cubename.npz}
        """        
        self.get_paths(pred_dir, targ_dir)

    def get_paths(self, pred_dir: str, targ_dir: str):
        """Match paths of target cubes with predicted cubes

        Each target cube gets 1 or more predicted cubes.

        Args:
            pred_dir (str): Directory with predictions, format is one of {pred_dir/tile/cubename.npz, pred_dir/tile/experiment_cubename.npz}
            targ_dir (str): Directory with targets, format is one of {targ_dir/target/tile/target_cubename.npz, targ_dir/target/tile/cubename.npz, targ_dir/tile/target_cubename.npz, targ_dir/tile/cubename.npz}
        """        
        print("Initializing filepaths...")

        pred_dir, targ_dir = Path(pred_dir), Path(targ_dir)

        if "target" in [d.name for d in targ_dir.glob("*") if d.is_dir()]:
            targ_dir = targ_dir/"target"

        assert({d.name for d in pred_dir.glob("*") if d.is_dir()}.issubset({d.name for d in targ_dir.glob("*") if d.is_dir()}))

        targ_paths = sorted(list(targ_dir.glob("**/*.npz")))

        filepaths = []
        for targ_path in tqdm(targ_paths):

            pred_paths = sorted(list(pred_dir.glob(f"**/*{self.__name_getter(targ_path)}")))
            assert (len(pred_paths) <= 10),"EarthNetScore is calculated with up to 10 predictions for a target, but more than 10 predictions were found."
            for pred_path in pred_paths:
                filepaths.append({"pred_filepath": pred_path, "targ_filepath": targ_path})
        
        self.filepaths = filepaths

        print("Filepaths initialized.")
        
    def __name_getter(self, path: Path) -> str:
        """Helper function gets Cubename from a Path

        Args:
            path (Path): One of Path/to/cubename.npz and Path/to/experiment_cubename.npz

        Returns:
            [str]: cubename (has format tile_startyear_startmonth_startday_endyear_endmonth_endday_hrxmin_hrxmax_hrymin_hrymax_mesoxmin_mesoxmax_mesoymin_mesoymax.npz)
        """        
        components = path.name.split("_")
        regex = re.compile('\d{2}[A-Z]{3}')
        if bool(regex.match(components[0])):
            return path.name
        else:
            assert(bool(regex.match(components[1])))
            return "_".join(components[1:]) 

    def compute_scores(self, n_workers: Optional[int] = -1) -> dict:
        """Compute subscores for all cubepaths

        Args:
            n_workers (Optional[int], optional): Number of workers, if -1 uses all CPUs, if 0 uses no multiprocessing. Defaults to -1.

        Returns:
            dict: data of format {cubename: score_dict}
        """        
        if n_workers == 0:
            all_scores = []
            print("Iteratively computing components for EarthNetScore")
            for filepaths in tqdm(self.filepaths):
                all_scores.append(CubeCalculator.get_scores(filepaths))
        else:
            if n_workers == -1:
                n_workers = multiprocessing.cpu_count()
            print(f"Computing components for EarthNetScore using {n_workers} processes")
            with multiprocessing.Pool(n_workers) as p:
                all_scores = list(tqdm(p.imap(CubeCalculator.get_scores, self.filepaths), total = len(self.filepaths)))

        data = {}
        for scores in all_scores:
            name = self.__name_getter(Path(scores["targ_filepath"])) 
            if name not in data:
                data[name] = [scores]
            else:
                data[name].append(scores)

        self.data = data

        print("Done computing scores.")

        return data

    def save_scores(self, output_file: str):
        """Save all subscores and debugging info as JSON

        Args:
            output_file (str): Output filepath, recommended to end with .json
        """        
        print("Saving data...")
        Path(output_file).parent.mkdir(parents=True, exist_ok=True)
        with open(output_file, "w") as fp:
            json.dump(self.data, fp)   
        print(f"Saved data to {output_file}.")
    
    def summarize(self, output_file: Optional[str] = None) -> Tuple[float, float, float, float, float]:
        """Calculate EarthNetScore from subscores and optionally save to file as JSON

        Args:
            output_file (Optional[str], optional): If not None, saves EarthNetScore to this path, recommended to end with .json. Defaults to None.

        Returns:
            Tuple[float, float, float, float, float]: ens, mad, ols, emd, ssim
        """        
        print("Calculating Earth Net Score...")
        scores = []
        for cube in tqdm(self.data):
            best_sample = self.__get_best_sample(self.data[cube])
            scores.append([best_sample["MAD"],best_sample["OLS"],best_sample["EMD"],best_sample["SSIM"]])
        scores = np.array(scores, dtype = np.float64)
        mean_scores = np.nanmean(scores, axis = 0).tolist()
        ens = self.__harmonic_mean(mean_scores)

        if output_file is not None:
            output_dict = {
                "EarthNetScore": ens,
                "Value (MAD)": mean_scores[0],
                "Trend (OLS)": mean_scores[1],
                "Distribution (EMD)": mean_scores[2],
                "Perceptual (SSIM)": mean_scores[3]
            }
            Path(output_file).parent.mkdir(parents=True, exist_ok=True)
            with open(output_file, "w") as fp:
                json.dump(output_dict, fp)

        print(f"EarthNetScore: {ens}\t Value (MAD): {mean_scores[0]}\t Trend (OLS): {mean_scores[1]}\t Distribution (EMD): {mean_scores[2]}\t Perceptual (SSIM): {mean_scores[3]}")
        return [ens]+mean_scores

    def __harmonic_mean(self, vals: Sequence[float]) -> Union[float, None]:
        """

        Calculates the harmonic mean of a list of values, safe for NaNs

        Args:
            vals (list): List of Floats
        Returns:
            float: harmonic mean
        """        
        vals = list(filter(None, vals))
        if len(vals) == 0:
            return None
        else:
            return min(1,len(vals)/sum([1/(v+1e-8) for v in vals]))
                
    def __get_best_sample(self, samples: Sequence[dict]) -> dict:
        """Gets best prediction out of 1 to n predictions. Safe to NaNs

        Args:
            samples (Sequence[dict]): List of dicts with subscores per sample

        Returns:
            dict: dict with subscores of best sample
        """        
        ens = np.array([self.__harmonic_mean([sample["MAD"],sample["OLS"],sample["EMD"],sample["SSIM"]]) for sample in samples], dtype = np.float64)

        try:
            min_idx = np.nanargmax(ens)
        except ValueError:
            min_idx = 0
        
        return samples[min_idx]
    
    @classmethod
    def get_ENS(cls, pred_dir: str, targ_dir: str, n_workers: Optional[int] = -1, data_output_file: Optional[str] = None, ens_output_file: Optional[str] = None):
        """Method to directly compute EarthNetScore

        Args:
            pred_dir (str): Directory with predictions, format is one of {pred_dir/tile/cubename.npz, pred_dir/tile/experiment_cubename.npz}
            targ_dir (str): Directory with targets, format is one of {targ_dir/target/tile/target_cubename.npz, targ_dir/target/tile/cubename.npz, targ_dir/tile/target_cubename.npz, targ_dir/tile/cubename.npz}
            n_workers (Optional[int], optional): Number of workers, if -1 uses all CPUs, if 0 uses no multiprocessing. Defaults to -1.
            data_output_file (Optional[str], optional): Output filepath for subscores and debugging information, recommended to end with .json. Defaults to None.
            ens_output_file (Optional[str], optional): Output filepath for EarthNetScore, recommended to end with .json. Defaults to None.
        """        

        self = cls(pred_dir, targ_dir)
        
        self.compute_scores(n_workers = n_workers)

        if data_output_file is not None:
            self.save_scores(output_file = data_output_file)
        
        self.summarize(output_file = ens_output_file)

if __name__=="__main__":

    parser = argparse.ArgumentParser(description="Parallel evaluation using EarthNetScore")
    parser.add_argument('--pred_dir', type = str, help='Path where Predictions are saved')
    parser.add_argument('--targ_dir', type = str, help ='Path where targets are saved')
    parser.add_argument('--data_output_file', type = str, help ='Filepath where output data will be saved')
    parser.add_argument('--ens_output_file', type = str, help ='Filepath where resulting EarthNetScore will be saved')

    args = parser.parse_args()
    start = time.time()
    EarthNetScore.get_ENS(args.pred_dir, args.targ_dir, data_output_file = args.data_output_file, ens_output_file = args.ens_output_file)
    end = time.time()

    print(f"Calculation done in {end - start} seconds.")



from copy import deepcopy
import numpy as np
import torch


def layout_to_in_out_slice(layout, in_len, out_len=None):
    t_axis = layout.find("T")
    num_axes = len(layout)
    in_slice = [slice(None, None), ] * num_axes
    out_slice = deepcopy(in_slice)
    in_slice[t_axis] = slice(None, in_len)
    if out_len is None:
        out_slice[t_axis] = slice(in_len, None)
    else:
        out_slice[t_axis] = slice(in_len, in_len + out_len)
    return in_slice, out_slice

def change_layout_np(data,
                     in_layout='NHWT', out_layout='NHWT',
                     ret_contiguous=False):
    # first convert to 'NHWT'
    if in_layout == 'NHWT':
        pass
    elif in_layout == 'NTHW':
        data = np.transpose(data,
                            axes=(0, 2, 3, 1))
    elif in_layout == 'NWHT':
        data = np.transpose(data,
                            axes=(0, 2, 1, 3))
    elif in_layout == 'NTCHW':
        data = data[:, :, 0, :, :]
        data = np.transpose(data,
                            axes=(0, 2, 3, 1))
    elif in_layout == 'NTHWC':
        data = data[:, :, :, :, 0]
        data = np.transpose(data,
                            axes=(0, 2, 3, 1))
    elif in_layout == 'NTWHC':
        data = data[:, :, :, :, 0]
        data = np.transpose(data,
                            axes=(0, 3, 2, 1))
    elif in_layout == 'TNHW':
        data = np.transpose(data,
                            axes=(1, 2, 3, 0))
    elif in_layout == 'TNCHW':
        data = data[:, :, 0, :, :]
        data = np.transpose(data,
                            axes=(1, 2, 3, 0))
    else:
        raise NotImplementedError

    if out_layout == 'NHWT':
        pass
    elif out_layout == 'NTHW':
        data = np.transpose(data,
                            axes=(0, 3, 1, 2))
    elif out_layout == 'NWHT':
        data = np.transpose(data,
                            axes=(0, 2, 1, 3))
    elif out_layout == 'NTCHW':
        data = np.transpose(data,
                            axes=(0, 3, 1, 2))
        data = np.expand_dims(data, axis=2)
    elif out_layout == 'NTHWC':
        data = np.transpose(data,
                            axes=(0, 3, 1, 2))
        data = np.expand_dims(data, axis=-1)
    elif out_layout == 'NTWHC':
        data = np.transpose(data,
                            axes=(0, 3, 2, 1))
        data = np.expand_dims(data, axis=-1)
    elif out_layout == 'TNHW':
        data = np.transpose(data,
                            axes=(3, 0, 1, 2))
    elif out_layout == 'TNCHW':
        data = np.transpose(data,
                            axes=(3, 0, 1, 2))
        data = np.expand_dims(data, axis=2)
    else:
        raise NotImplementedError
    if ret_contiguous:
        data = data.ascontiguousarray()
    return data

def change_layout_torch(data,
                        in_layout='NHWT', out_layout='NHWT',
                        ret_contiguous=False):
    # first convert to 'NHWT'
    if in_layout == 'NHWT':
        pass
    elif in_layout == 'NTHW':
        data = data.permute(0, 2, 3, 1)
    elif in_layout == 'NTCHW':
        data = data[:, :, 0, :, :]
        data = data.permute(0, 2, 3, 1)
    elif in_layout == 'NTHWC':
        data = data[:, :, :, :, 0]
        data = data.permute(0, 2, 3, 1)
    elif in_layout == 'TNHW':
        data = data.permute(1, 2, 3, 0)
    elif in_layout == 'TNCHW':
        data = data[:, :, 0, :, :]
        data = data.permute(1, 2, 3, 0)
    else:
        raise NotImplementedError

    if out_layout == 'NHWT':
        pass
    elif out_layout == 'NTHW':
        data = data.permute(0, 3, 1, 2)
    elif out_layout == 'NTCHW':
        data = data.permute(0, 3, 1, 2)
        data = torch.unsqueeze(data, dim=2)
    elif out_layout == 'NTHWC':
        data = data.permute(0, 3, 1, 2)
        data = torch.unsqueeze(data, dim=-1)
    elif out_layout == 'TNHW':
        data = data.permute(3, 0, 1, 2)
    elif out_layout == 'TNCHW':
        data = data.permute(3, 0, 1, 2)
        data = torch.unsqueeze(data, dim=2)
    else:
        raise NotImplementedError
    if ret_contiguous:
        data = data.contiguous()
    return data



import os
import warnings
from typing import Callable, Dict, Optional, Union, List
import torch
from pytorch_lightning.utilities.cloud_io import load as pl_load


def average_checkpoints(checkpoint_paths: List[str] = None):
    r""" Code is adapted from https://github.com/awslabs/autogluon/blob/a818545e047f2bcda13569568a7fd611abdfc876/multimodal/src/autogluon/multimodal/utils/checkpoint.py#L13

    Average a list of checkpoints' state_dicts.
    Reference: https://github.com/rwightman/pytorch-image-models/blob/master/avg_checkpoints.py
    Parameters
    ----------
    checkpoint_paths
        A list of model checkpoint paths.
    Returns
    -------
    The averaged state_dict.
    """
    if len(checkpoint_paths) > 1:
        avg_state_dict = {}
        avg_counts = {}
        for per_path in checkpoint_paths:
            state_dict = torch.load(per_path, map_location=torch.device("cpu"))["state_dict"]
            for k, v in state_dict.items():
                if k not in avg_state_dict:
                    avg_state_dict[k] = v.clone().to(dtype=torch.float64)
                    avg_counts[k] = 1
                else:
                    avg_state_dict[k] += v.to(dtype=torch.float64)
                    avg_counts[k] += 1
            del state_dict

        for k, v in avg_state_dict.items():
            v.div_(avg_counts[k])

        # convert to float32.
        float32_info = torch.finfo(torch.float32)
        for k in avg_state_dict:
            avg_state_dict[k].clamp_(float32_info.min, float32_info.max).to(dtype=torch.float32)
    else:
        avg_state_dict = torch.load(checkpoint_paths[0], map_location=torch.device("cpu"))["state_dict"]

    return avg_state_dict

def average_pl_checkpoints(pl_checkpoint_paths: List[str] = None, delete_prefix_len: int = len("")):
    r""" Code is adapted from https://github.com/awslabs/autogluon/blob/a818545e047f2bcda13569568a7fd611abdfc876/multimodal/src/autogluon/multimodal/utils/checkpoint.py#L13

    Average a list of checkpoints' state_dicts.
    Reference: https://github.com/rwightman/pytorch-image-models/blob/master/avg_checkpoints.py
    Parameters
    ----------
    checkpoint_paths
        A list of model checkpoint paths.
    Returns
    -------
    The averaged state_dict.
    """
    if len(pl_checkpoint_paths) > 1:
        avg_state_dict = {}
        avg_counts = {}
        for per_path in pl_checkpoint_paths:
            state_dict = pl_ckpt_to_pytorch_state_dict(per_path,
                                                       map_location=torch.device("cpu"),
                                                       delete_prefix_len=delete_prefix_len)
            for k, v in state_dict.items():
                if k not in avg_state_dict:
                    avg_state_dict[k] = v.clone().to(dtype=torch.float64)
                    avg_counts[k] = 1
                else:
                    avg_state_dict[k] += v.to(dtype=torch.float64)
                    avg_counts[k] += 1
            del state_dict

        for k, v in avg_state_dict.items():
            v.div_(avg_counts[k])

        # convert to float32.
        float32_info = torch.finfo(torch.float32)
        for k in avg_state_dict:
            avg_state_dict[k].clamp_(float32_info.min, float32_info.max).to(dtype=torch.float32)
    else:
        avg_state_dict = pl_ckpt_to_pytorch_state_dict(pl_checkpoint_paths[0],
                                                       map_location=torch.device("cpu"),
                                                       delete_prefix_len=delete_prefix_len)

    return avg_state_dict

def pl_ckpt_to_pytorch_state_dict(
        checkpoint_path: str,
        map_location: Optional[Union[Dict[str, str], str, torch.device, int, Callable]] = None,
        delete_prefix_len: int = len("")):
    r"""
    Parameters
    ----------
    checkpoint_path:    str
    map_location
        A function, torch.device, string or a dict specifying how to remap storage locations.
        The same as the arg `map_location` in `torch.load()`.
    delete_prefix_len:  int
        Delete the first several characters in the keys of state_dict.

    Returns
    -------
    pytorch_state_dict: OrderedDict
    """
    if map_location is not None:
        checkpoint = pl_load(checkpoint_path, map_location=map_location)
    else:
        checkpoint = pl_load(checkpoint_path, map_location=lambda storage, loc: storage)
    pl_ckpt_state_dict = checkpoint["state_dict"]
    pytorch_state_dict = {key[delete_prefix_len:]: val
                          for key, val in pl_ckpt_state_dict.items()}
    return pytorch_state_dict


def s3_download_pretrained_ckpt(ckpt_name, save_dir=None, exist_ok=False):
    if save_dir is None:
        from ..config import cfg
        save_dir = cfg.pretrained_checkpoints_dir
    if os.path.exists(os.path.join(save_dir, ckpt_name)) and not exist_ok:
        warnings.warn(f"Checkpoint file {os.path.join(save_dir, ckpt_name)} already exists!")
    else:
        os.makedirs(save_dir, exist_ok=True)
        os.system(f"aws s3 cp --no-sign-request s3://earthformer/pretrained_checkpoints/{ckpt_name} "
                  f"{save_dir}")



import warnings
from typing import Union, Dict
from shutil import copyfile
from copy import deepcopy
import inspect
import pickle
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.optim.lr_scheduler import LambdaLR, CosineAnnealingLR
import torchmetrics
import pytorch_lightning as pl
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor, DeviceStatsMonitor, Callback
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from omegaconf import OmegaConf
import os
import argparse
from einops import rearrange
from pytorch_lightning import Trainer, seed_everything
from earthformer.config import cfg
from earthformer.utils.optim import SequentialLR, warmup_lambda
from earthformer.utils.utils import get_parameter_names
from earthformer.utils.checkpoint import pl_ckpt_to_pytorch_state_dict, s3_download_pretrained_ckpt
from earthformer.utils.layout import layout_to_in_out_slice
from earthformer.visualization.sevir.sevir_vis_seq import save_example_vis_results
from earthformer.metrics.sevir import SEVIRSkillScore
from earthformer.cuboid_transformer.cuboid_transformer import CuboidTransformerModel
from earthformer.datasets.sevir.sevir_torch_wrap import SEVIRLightningDataModule
from earthformer.utils.apex_ddp import ApexDDPStrategy


_curr_dir = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))
exps_dir = os.path.join(_curr_dir, "experiments")
pretrained_checkpoints_dir = cfg.pretrained_checkpoints_dir
pytorch_state_dict_name = "earthformer_sevir.pt"

class CuboidSEVIRPLModule(pl.LightningModule):

    def __init__(self,
                 total_num_steps: int,
                 oc_file: str = None,
                 save_dir: str = None):
        super(CuboidSEVIRPLModule, self).__init__()
        self._max_train_iter = total_num_steps
        if oc_file is not None:
            oc_from_file = OmegaConf.load(open(oc_file, "r"))
        else:
            oc_from_file = None
        oc = self.get_base_config(oc_from_file=oc_from_file)
        model_cfg = OmegaConf.to_object(oc.model)
        num_blocks = len(model_cfg["enc_depth"])
        if isinstance(model_cfg["self_pattern"], str):
            enc_attn_patterns = [model_cfg["self_pattern"]] * num_blocks
        else:
            enc_attn_patterns = OmegaConf.to_container(model_cfg["self_pattern"])
        if isinstance(model_cfg["cross_self_pattern"], str):
            dec_self_attn_patterns = [model_cfg["cross_self_pattern"]] * num_blocks
        else:
            dec_self_attn_patterns = OmegaConf.to_container(model_cfg["cross_self_pattern"])
        if isinstance(model_cfg["cross_pattern"], str):
            dec_cross_attn_patterns = [model_cfg["cross_pattern"]] * num_blocks
        else:
            dec_cross_attn_patterns = OmegaConf.to_container(model_cfg["cross_pattern"])

        self.torch_nn_module = CuboidTransformerModel(
            input_shape=model_cfg["input_shape"],
            target_shape=model_cfg["target_shape"],
            base_units=model_cfg["base_units"],
            block_units=model_cfg["block_units"],
            scale_alpha=model_cfg["scale_alpha"],
            enc_depth=model_cfg["enc_depth"],
            dec_depth=model_cfg["dec_depth"],
            enc_use_inter_ffn=model_cfg["enc_use_inter_ffn"],
            dec_use_inter_ffn=model_cfg["dec_use_inter_ffn"],
            dec_hierarchical_pos_embed=model_cfg["dec_hierarchical_pos_embed"],
            downsample=model_cfg["downsample"],
            downsample_type=model_cfg["downsample_type"],
            enc_attn_patterns=enc_attn_patterns,
            dec_self_attn_patterns=dec_self_attn_patterns,
            dec_cross_attn_patterns=dec_cross_attn_patterns,
            dec_cross_last_n_frames=model_cfg["dec_cross_last_n_frames"],
            dec_use_first_self_attn=model_cfg["dec_use_first_self_attn"],
            num_heads=model_cfg["num_heads"],
            attn_drop=model_cfg["attn_drop"],
            proj_drop=model_cfg["proj_drop"],
            ffn_drop=model_cfg["ffn_drop"],
            upsample_type=model_cfg["upsample_type"],
            ffn_activation=model_cfg["ffn_activation"],
            gated_ffn=model_cfg["gated_ffn"],
            norm_layer=model_cfg["norm_layer"],
            # global vectors
            num_global_vectors=model_cfg["num_global_vectors"],
            use_dec_self_global=model_cfg["use_dec_self_global"],
            dec_self_update_global=model_cfg["dec_self_update_global"],
            use_dec_cross_global=model_cfg["use_dec_cross_global"],
            use_global_vector_ffn=model_cfg["use_global_vector_ffn"],
            use_global_self_attn=model_cfg["use_global_self_attn"],
            separate_global_qkv=model_cfg["separate_global_qkv"],
            global_dim_ratio=model_cfg["global_dim_ratio"],
            # initial_downsample
            initial_downsample_type=model_cfg["initial_downsample_type"],
            initial_downsample_activation=model_cfg["initial_downsample_activation"],
            # initial_downsample_type=="stack_conv"
            initial_downsample_stack_conv_num_layers=model_cfg["initial_downsample_stack_conv_num_layers"],
            initial_downsample_stack_conv_dim_list=model_cfg["initial_downsample_stack_conv_dim_list"],
            initial_downsample_stack_conv_downscale_list=model_cfg["initial_downsample_stack_conv_downscale_list"],
            initial_downsample_stack_conv_num_conv_list=model_cfg["initial_downsample_stack_conv_num_conv_list"],
            # misc
            padding_type=model_cfg["padding_type"],
            z_init_method=model_cfg["z_init_method"],
            checkpoint_level=model_cfg["checkpoint_level"],
            pos_embed_type=model_cfg["pos_embed_type"],
            use_relative_pos=model_cfg["use_relative_pos"],
            self_attn_use_final_proj=model_cfg["self_attn_use_final_proj"],
            # initialization
            attn_linear_init_mode=model_cfg["attn_linear_init_mode"],
            ffn_linear_init_mode=model_cfg["ffn_linear_init_mode"],
            conv_init_mode=model_cfg["conv_init_mode"],
            down_up_linear_init_mode=model_cfg["down_up_linear_init_mode"],
            norm_init_mode=model_cfg["norm_init_mode"],
        )

        self.total_num_steps = total_num_steps
        if oc_file is not None:
            oc_from_file = OmegaConf.load(open(oc_file, "r"))
        else:
            oc_from_file = None
        oc = self.get_base_config(oc_from_file=oc_from_file)
        self.save_hyperparameters(oc)
        self.oc = oc
        # layout
        self.in_len = oc.layout.in_len
        self.out_len = oc.layout.out_len
        self.layout = oc.layout.layout
        # optimization
        self.max_epochs = oc.optim.max_epochs
        self.optim_method = oc.optim.method
        self.lr = oc.optim.lr
        self.wd = oc.optim.wd
        # lr_scheduler
        self.total_num_steps = total_num_steps
        self.lr_scheduler_mode = oc.optim.lr_scheduler_mode
        self.warmup_percentage = oc.optim.warmup_percentage
        self.min_lr_ratio = oc.optim.min_lr_ratio
        # logging
        self.save_dir = save_dir
        self.logging_prefix = oc.logging.logging_prefix
        # visualization
        self.train_example_data_idx_list = list(oc.vis.train_example_data_idx_list)
        self.val_example_data_idx_list = list(oc.vis.val_example_data_idx_list)
        self.test_example_data_idx_list = list(oc.vis.test_example_data_idx_list)
        self.eval_example_only = oc.vis.eval_example_only

        self.configure_save(cfg_file_path=oc_file)
        # evaluation
        self.metrics_list = oc.dataset.metrics_list
        self.threshold_list = oc.dataset.threshold_list
        self.metrics_mode = oc.dataset.metrics_mode
        self.valid_mse = torchmetrics.MeanSquaredError()
        self.valid_mae = torchmetrics.MeanAbsoluteError()
        self.valid_score = SEVIRSkillScore(
            mode=self.metrics_mode,
            seq_len=self.out_len,
            layout=self.layout,
            threshold_list=self.threshold_list,
            metrics_list=self.metrics_list,
            eps=1e-4,)
        self.test_mse = torchmetrics.MeanSquaredError()
        self.test_mae = torchmetrics.MeanAbsoluteError()
        self.test_score = SEVIRSkillScore(
            mode=self.metrics_mode,
            seq_len=self.out_len,
            layout=self.layout,
            threshold_list=self.threshold_list,
            metrics_list=self.metrics_list,
            eps=1e-4,)

    def configure_save(self, cfg_file_path=None):
        self.save_dir = os.path.join(exps_dir, self.save_dir)
        os.makedirs(self.save_dir, exist_ok=True)
        self.scores_dir = os.path.join(self.save_dir, 'scores')
        os.makedirs(self.scores_dir, exist_ok=True)
        if cfg_file_path is not None:
            cfg_file_target_path = os.path.join(self.save_dir, "cfg.yaml")
            if (not os.path.exists(cfg_file_target_path)) or \
                    (not os.path.samefile(cfg_file_path, cfg_file_target_path)):
                copyfile(cfg_file_path, cfg_file_target_path)
        self.example_save_dir = os.path.join(self.save_dir, "examples")
        os.makedirs(self.example_save_dir, exist_ok=True)

    def get_base_config(self, oc_from_file=None):
        oc = OmegaConf.create()
        oc.dataset = self.get_dataset_config()
        oc.layout = self.get_layout_config()
        oc.optim = self.get_optim_config()
        oc.logging = self.get_logging_config()
        oc.trainer = self.get_trainer_config()
        oc.vis = self.get_vis_config()
        oc.model = self.get_model_config()
        if oc_from_file is not None:
            oc = OmegaConf.merge(oc, oc_from_file)
        return oc

    @staticmethod
    def get_dataset_config():
        oc = OmegaConf.create()
        oc.dataset_name = "sevir"
        oc.img_height = 384
        oc.img_width = 384
        oc.in_len = 13
        oc.out_len = 12
        oc.seq_len = 25
        oc.plot_stride = 2
        oc.interval_real_time = 5
        oc.sample_mode = "sequent"
        oc.stride = oc.out_len
        oc.layout = "NTHWC"
        oc.start_date = None
        oc.train_val_split_date = (2019, 1, 1)
        oc.train_test_split_date = (2019, 6, 1)
        oc.end_date = None
        oc.metrics_mode = "0"
        oc.metrics_list = ('csi', 'pod', 'sucr', 'bias')
        oc.threshold_list = (16, 74, 133, 160, 181, 219)
        return oc
    
    @classmethod
    def get_model_config(cls):
        cfg = OmegaConf.create()
        dataset_oc = cls.get_dataset_config()
        height = dataset_oc.img_height
        width = dataset_oc.img_width
        in_len = dataset_oc.in_len
        out_len = dataset_oc.out_len
        data_channels = 1
        cfg.input_shape = (in_len, height, width, data_channels)
        cfg.target_shape = (out_len, height, width, data_channels)

        cfg.base_units = 64
        cfg.block_units = None # multiply by 2 when downsampling in each layer
        cfg.scale_alpha = 1.0

        cfg.enc_depth = [1, 1]
        cfg.dec_depth = [1, 1]
        cfg.enc_use_inter_ffn = True
        cfg.dec_use_inter_ffn = True
        cfg.dec_hierarchical_pos_embed = True

        cfg.downsample = 2
        cfg.downsample_type = "patch_merge"
        cfg.upsample_type = "upsample"

        cfg.num_global_vectors = 8
        cfg.use_dec_self_global = True
        cfg.dec_self_update_global = True
        cfg.use_dec_cross_global = True
        cfg.use_global_vector_ffn = True
        cfg.use_global_self_attn = False
        cfg.separate_global_qkv = False
        cfg.global_dim_ratio = 1

        cfg.self_pattern = 'axial'
        cfg.cross_self_pattern = 'axial'
        cfg.cross_pattern = 'cross_1x1'
        cfg.dec_cross_last_n_frames = None

        cfg.attn_drop = 0.1
        cfg.proj_drop = 0.1
        cfg.ffn_drop = 0.1
        cfg.num_heads = 4

        cfg.ffn_activation = 'gelu'
        cfg.gated_ffn = False
        cfg.norm_layer = 'layer_norm'
        cfg.padding_type = 'zeros'
        cfg.pos_embed_type = "t+hw"
        cfg.use_relative_pos = True
        cfg.self_attn_use_final_proj = True
        cfg.dec_use_first_self_attn = False

        cfg.z_init_method = 'zeros'
        cfg.checkpoint_level = 2
        # initial downsample and final upsample
        cfg.initial_downsample_type = "stack_conv"
        cfg.initial_downsample_activation = "leaky"
        cfg.initial_downsample_stack_conv_num_layers = 3
        cfg.initial_downsample_stack_conv_dim_list = [4, 16, cfg.base_units]
        cfg.initial_downsample_stack_conv_downscale_list = [3, 2, 2]
        cfg.initial_downsample_stack_conv_num_conv_list = [2, 2, 2]
        # initialization
        cfg.attn_linear_init_mode = "0"
        cfg.ffn_linear_init_mode = "0"
        cfg.conv_init_mode = "0"
        cfg.down_up_linear_init_mode = "0"
        cfg.norm_init_mode = "0"
        return cfg

    @classmethod
    def get_layout_config(cls):
        oc = OmegaConf.create()
        dataset_oc = cls.get_dataset_config()
        oc.in_len = dataset_oc.in_len
        oc.out_len = dataset_oc.out_len
        oc.layout = dataset_oc.layout
        return oc

    @staticmethod
    def get_optim_config():
        oc = OmegaConf.create()
        oc.seed = None
        oc.total_batch_size = 32
        oc.micro_batch_size = 8

        oc.method = "adamw"
        oc.lr = 1E-3
        oc.wd = 1E-5
        oc.gradient_clip_val = 1.0
        oc.max_epochs = 100
        # scheduler
        oc.warmup_percentage = 0.2
        oc.lr_scheduler_mode = "cosine"  # Can be strings like 'linear', 'cosine', 'platue'
        oc.min_lr_ratio = 0.1
        oc.warmup_min_lr_ratio = 0.1
        # early stopping
        oc.early_stop = False
        oc.early_stop_mode = "min"
        oc.early_stop_patience = 20
        oc.save_top_k = 1
        return oc

    @staticmethod
    def get_logging_config():
        oc = OmegaConf.create()
        oc.logging_prefix = "SEVIR"
        oc.monitor_lr = True
        oc.monitor_device = False
        oc.track_grad_norm = -1
        oc.use_wandb = True
        return oc

    @staticmethod
    def get_trainer_config():
        oc = OmegaConf.create()
        oc.check_val_every_n_epoch = 1
        oc.log_step_ratio = 0.001  # Logging every 1% of the total training steps per epoch
        oc.precision = 32
        return oc

    @classmethod
    def get_vis_config(cls):
        oc = OmegaConf.create()
        dataset_oc = cls.get_dataset_config()
        oc.train_example_data_idx_list = [0, ]
        oc.val_example_data_idx_list = [80, ]
        oc.test_example_data_idx_list = [0, 80, 160, 240, 320, 400]
        oc.eval_example_only = False
        oc.plot_stride = dataset_oc.plot_stride
        return oc

    def configure_optimizers(self):
        # Configure the optimizer. Disable the weight decay for layer norm weights and all bias terms.
        decay_parameters = get_parameter_names(self.torch_nn_module, [nn.LayerNorm])
        decay_parameters = [name for name in decay_parameters if "bias" not in name]
        optimizer_grouped_parameters = [{
            'params': [p for n, p in self.torch_nn_module.named_parameters()
                       if n in decay_parameters],
            'weight_decay': self.oc.optim.wd
        }, {
            'params': [p for n, p in self.torch_nn_module.named_parameters()
                       if n not in decay_parameters],
            'weight_decay': 0.0
        }]

        if self.oc.optim.method == 'adamw':
            optimizer = torch.optim.AdamW(params=optimizer_grouped_parameters,
                                          lr=self.oc.optim.lr,
                                          weight_decay=self.oc.optim.wd)
        else:
            raise NotImplementedError

        warmup_iter = int(np.round(self.oc.optim.warmup_percentage * self.total_num_steps))

        if self.oc.optim.lr_scheduler_mode == 'cosine':
            warmup_scheduler = LambdaLR(optimizer,
                                        lr_lambda=warmup_lambda(warmup_steps=warmup_iter,
                                                                min_lr_ratio=self.oc.optim.warmup_min_lr_ratio))
            cosine_scheduler = CosineAnnealingLR(optimizer,
                                                 T_max=(self.total_num_steps - warmup_iter),
                                                 eta_min=self.oc.optim.min_lr_ratio * self.oc.optim.lr)
            lr_scheduler = SequentialLR(optimizer, schedulers=[warmup_scheduler, cosine_scheduler],
                                        milestones=[warmup_iter])
            lr_scheduler_config = {
                'scheduler': lr_scheduler,
                'interval': 'step',
                'frequency': 1,
            }
        else:
            raise NotImplementedError
        return {'optimizer': optimizer, 'lr_scheduler': lr_scheduler_config}

    def set_trainer_kwargs(self, **kwargs):
        r"""
        Default kwargs used when initializing pl.Trainer
        """
        checkpoint_callback = ModelCheckpoint(
            monitor="valid_loss_epoch",
            dirpath=os.path.join(self.save_dir, "checkpoints"),
            filename="model-{epoch:03d}",
            save_top_k=self.oc.optim.save_top_k,
            save_last=True,
            mode="min",
        )
        callbacks = kwargs.pop("callbacks", [])
        assert isinstance(callbacks, list)
        for ele in callbacks:
            assert isinstance(ele, Callback)
        callbacks += [checkpoint_callback, ]
        if self.oc.logging.monitor_lr:
            callbacks += [LearningRateMonitor(logging_interval='step'), ]
        if self.oc.logging.monitor_device:
            callbacks += [DeviceStatsMonitor(), ]
        if self.oc.optim.early_stop:
            callbacks += [EarlyStopping(monitor="valid_loss_epoch",
                                        min_delta=0.0,
                                        patience=self.oc.optim.early_stop_patience,
                                        verbose=False,
                                        mode=self.oc.optim.early_stop_mode), ]

        logger = kwargs.pop("logger", [])
        tb_logger = pl_loggers.TensorBoardLogger(save_dir=self.save_dir)
        csv_logger = pl_loggers.CSVLogger(save_dir=self.save_dir)
        logger += [tb_logger, csv_logger]
        if self.oc.logging.use_wandb:
            wandb_logger = pl_loggers.WandbLogger(project=self.oc.logging.logging_prefix,
                                                  save_dir=self.save_dir)
            logger += [wandb_logger, ]

        log_every_n_steps = max(1, int(self.oc.trainer.log_step_ratio * self.total_num_steps))
        trainer_init_keys = inspect.signature(Trainer).parameters.keys()
        ret = dict(
            callbacks=callbacks,
            # log
            logger=logger,
            log_every_n_steps=log_every_n_steps,
            track_grad_norm=self.oc.logging.track_grad_norm,
            # save
            default_root_dir=self.save_dir,
            # ddp
            accelerator="gpu",
            # strategy="ddp",
            strategy=ApexDDPStrategy(find_unused_parameters=False, delay_allreduce=True),
            # optimization
            max_epochs=self.oc.optim.max_epochs,
            check_val_every_n_epoch=self.oc.trainer.check_val_every_n_epoch,
            gradient_clip_val=self.oc.optim.gradient_clip_val,
            # NVIDIA amp
            precision=self.oc.trainer.precision,
        )
        oc_trainer_kwargs = OmegaConf.to_object(self.oc.trainer)
        oc_trainer_kwargs = {key: val for key, val in oc_trainer_kwargs.items() if key in trainer_init_keys}
        ret.update(oc_trainer_kwargs)
        ret.update(kwargs)
        return ret

    @classmethod
    def get_total_num_steps(
            cls,
            num_samples: int,
            total_batch_size: int,
            epoch: int = None):
        r"""
        Parameters
        ----------
        num_samples:    int
            The number of samples of the datasets. `num_samples / micro_batch_size` is the number of steps per epoch.
        total_batch_size:   int
            `total_batch_size == micro_batch_size * world_size * grad_accum`
        """
        if epoch is None:
            epoch = cls.get_optim_config().max_epochs
        return int(epoch * num_samples / total_batch_size)

    @staticmethod
    def get_sevir_datamodule(dataset_oc,
                             micro_batch_size: int = 1,
                             num_workers: int = 8):
        dm = SEVIRLightningDataModule(
            seq_len=dataset_oc["seq_len"],
            sample_mode=dataset_oc["sample_mode"],
            stride=dataset_oc["stride"],
            batch_size=micro_batch_size,
            layout=dataset_oc["layout"],
            output_type=np.float32,
            preprocess=True,
            rescale_method="01",
            verbose=False,
            # datamodule_only
            dataset_name=dataset_oc["dataset_name"],
            start_date=dataset_oc["start_date"],
            train_val_split_date=dataset_oc["train_val_split_date"],
            train_test_split_date=dataset_oc["train_test_split_date"],
            end_date=dataset_oc["end_date"],
            num_workers=num_workers,)
        return dm

    @property
    def in_slice(self):
        if not hasattr(self, "_in_slice"):
            in_slice, out_slice = layout_to_in_out_slice(layout=self.layout,
                                                         in_len=self.in_len,
                                                         out_len=self.out_len)
            self._in_slice = in_slice
            self._out_slice = out_slice
        return self._in_slice

    @property
    def out_slice(self):
        if not hasattr(self, "_out_slice"):
            in_slice, out_slice = layout_to_in_out_slice(layout=self.layout,
                                                         in_len=self.in_len,
                                                         out_len=self.out_len)
            self._in_slice = in_slice
            self._out_slice = out_slice
        return self._out_slice

    def forward(self, in_seq, out_seq):
        output = self.torch_nn_module(in_seq)
        loss = F.mse_loss(output, out_seq)
        return output, loss

    def training_step(self, batch, batch_idx):
        data_seq = batch['vil'].contiguous()
        x = data_seq[self.in_slice]
        y = data_seq[self.out_slice]
        y_hat, loss = self(x, y)
        micro_batch_size = x.shape[self.layout.find("N")]
        data_idx = int(batch_idx * micro_batch_size)
        self.save_vis_step_end(
            data_idx=data_idx,
            in_seq=x,
            target_seq=y,
            pred_seq=y_hat,
            mode="train"
        )
        self.log('train_loss', loss,
                 on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx):
        data_seq = batch['vil'].contiguous()
        x = data_seq[self.in_slice]
        y = data_seq[self.out_slice]
        micro_batch_size = x.shape[self.layout.find("N")]
        data_idx = int(batch_idx * micro_batch_size)
        if not self.eval_example_only or data_idx in self.val_example_data_idx_list:
            y_hat, _ = self(x, y)
            self.save_vis_step_end(
                data_idx=data_idx,
                in_seq=x,
                target_seq=y,
                pred_seq=y_hat,
                mode="val"
            )
            if self.precision == 16:
                y_hat = y_hat.float()
            step_mse = self.valid_mse(y_hat, y)
            step_mae = self.valid_mae(y_hat, y)
            self.valid_score.update(y_hat, y)
            self.log('valid_frame_mse_step', step_mse,
                     prog_bar=True, on_step=True, on_epoch=False)
            self.log('valid_frame_mae_step', step_mae,
                     prog_bar=True, on_step=True, on_epoch=False)
        return None

    def validation_epoch_end(self, outputs):
        valid_mse = self.valid_mse.compute()
        valid_mae = self.valid_mae.compute()
        self.log('valid_frame_mse_epoch', valid_mse,
                 prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_frame_mae_epoch', valid_mae,
                 prog_bar=True, on_step=False, on_epoch=True)
        self.valid_mse.reset()
        self.valid_mae.reset()
        valid_score = self.valid_score.compute()
        self.log("valid_loss_epoch", -valid_score["avg"]["csi"],
                 prog_bar=True, on_step=False, on_epoch=True)
        self.log_score_epoch_end(score_dict=valid_score, mode="val")
        self.valid_score.reset()
        self.save_score_epoch_end(score_dict=valid_score,
                                  mse=valid_mse,
                                  mae=valid_mae,
                                  mode="val")

    def test_step(self, batch, batch_idx):
        data_seq = batch['vil'].contiguous()
        x = data_seq[self.in_slice]
        y = data_seq[self.out_slice]
        micro_batch_size = x.shape[self.layout.find("N")]
        data_idx = int(batch_idx * micro_batch_size)
        if not self.eval_example_only or data_idx in self.test_example_data_idx_list:
            y_hat, _ = self(x, y)
            self.save_vis_step_end(
                data_idx=data_idx,
                in_seq=x,
                target_seq=y,
                pred_seq=y_hat,
                mode="test"
            )
            if self.precision == 16:
                y_hat = y_hat.float()
            step_mse = self.test_mse(y_hat, y)
            step_mae = self.test_mae(y_hat, y)
            self.test_score.update(y_hat, y)
            self.log('test_frame_mse_step', step_mse,
                     prog_bar=True, on_step=True, on_epoch=False)
            self.log('test_frame_mae_step', step_mae,
                     prog_bar=True, on_step=True, on_epoch=False)
        return None

    def test_epoch_end(self, outputs):
        test_mse = self.test_mse.compute()
        test_mae = self.test_mae.compute()
        self.log('test_frame_mse_epoch', test_mse,
                 prog_bar=True, on_step=False, on_epoch=True)
        self.log('test_frame_mae_epoch', test_mae,
                 prog_bar=True, on_step=False, on_epoch=True)
        self.test_mse.reset()
        self.test_mae.reset()
        test_score = self.test_score.compute()
        self.log_score_epoch_end(score_dict=test_score, mode="test")
        self.test_score.reset()
        self.save_score_epoch_end(score_dict=test_score,
                                  mse=test_mse,
                                  mae=test_mae,
                                  mode="test")

    def log_score_epoch_end(self, score_dict: Dict, mode: str = "val"):
        if mode == "val":
            log_mode_prefix = "valid"
        elif mode == "test":
            log_mode_prefix = "test"
        else:
            raise ValueError(f"Wrong mode {mode}. Must be 'val' or 'test'.")
        for metrics in self.metrics_list:
            for thresh in self.threshold_list:
                score_mean = np.mean(score_dict[thresh][metrics]).item()
                self.log(f"{log_mode_prefix}_{metrics}_{thresh}_epoch", score_mean)
            score_avg_mean = score_dict.get("avg", None)
            if score_avg_mean is not None:
                score_avg_mean = np.mean(score_avg_mean[metrics]).item()
                self.log(f"{log_mode_prefix}_{metrics}_avg_epoch", score_avg_mean)

    def save_score_epoch_end(self,
                             score_dict: Dict,
                             mse: Union[np.ndarray, float],
                             mae: Union[np.ndarray, float],
                             mode: str = "val"):
        assert mode in ["val", "test"], f"Wrong mode {mode}. Must be 'val' or 'test'."
        if self.local_rank == 0:
            save_dict = deepcopy(score_dict)
            save_dict.update(dict(mse=mse, mae=mae))
            if self.scores_dir is not None:
                save_path = os.path.join(self.scores_dir, f"{mode}_results_epoch_{self.current_epoch}.pkl")
                f = open(save_path, 'wb')
                pickle.dump(save_dict, f)
                f.close()

    def save_vis_step_end(
            self,
            data_idx: int,
            in_seq: torch.Tensor,
            target_seq: torch.Tensor,
            pred_seq: torch.Tensor,
            mode: str = "train"):
        r"""
        Parameters
        ----------
        data_idx:   int
            data_idx == batch_idx * micro_batch_size
        """
        if self.local_rank == 0:
            if mode == "train":
                example_data_idx_list = self.train_example_data_idx_list
            elif mode == "val":
                example_data_idx_list = self.val_example_data_idx_list
            elif mode == "test":
                example_data_idx_list = self.test_example_data_idx_list
            else:
                raise ValueError(f"Wrong mode {mode}! Must be in ['train', 'val', 'test'].")
            if data_idx in example_data_idx_list:
                save_example_vis_results(
                    save_dir=self.example_save_dir,
                    save_prefix=f'{mode}_epoch_{self.current_epoch}_data_{data_idx}',
                    in_seq=in_seq.detach().float().cpu().numpy(),
                    target_seq=target_seq.detach().float().cpu().numpy(),
                    pred_seq=pred_seq.detach().float().cpu().numpy(),
                    layout=self.layout,
                    plot_stride=self.oc.vis.plot_stride,
                    label=self.oc.logging.logging_prefix,
                    interval_real_time=self.oc.dataset.interval_real_time)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--save', default='tmp_sevir', type=str)
    parser.add_argument('--gpus', default=1, type=int)
    parser.add_argument('--cfg', default=None, type=str)
    parser.add_argument('--test', action='store_true')
    parser.add_argument('--pretrained', action='store_true',
                        help='Load pretrained checkpoints for test.')
    parser.add_argument('--ckpt_name', default=None, type=str,
                        help='The model checkpoint trained on SEVIR.')
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.pretrained:
        args.cfg = os.path.abspath(os.path.join(os.path.dirname(__file__), "earthformer_sevir_v1.yaml"))
    if args.cfg is not None:
        oc_from_file = OmegaConf.load(open(args.cfg, "r"))
        dataset_oc = OmegaConf.to_object(oc_from_file.dataset)
        total_batch_size = oc_from_file.optim.total_batch_size
        micro_batch_size = oc_from_file.optim.micro_batch_size
        max_epochs = oc_from_file.optim.max_epochs
        seed = oc_from_file.optim.seed
    else:
        dataset_oc = OmegaConf.to_object(CuboidSEVIRPLModule.get_dataset_config())
        micro_batch_size = 1
        total_batch_size = int(micro_batch_size * args.gpus)
        max_epochs = None
        seed = 0
    seed_everything(seed, workers=True)
    dm = CuboidSEVIRPLModule.get_sevir_datamodule(
        dataset_oc=dataset_oc,
        micro_batch_size=micro_batch_size,
        num_workers=8,)
    dm.prepare_data()
    dm.setup()
    accumulate_grad_batches = total_batch_size // (micro_batch_size * args.gpus)
    total_num_steps = CuboidSEVIRPLModule.get_total_num_steps(
        epoch=max_epochs,
        num_samples=dm.num_train_samples,
        total_batch_size=total_batch_size,
    )
    pl_module = CuboidSEVIRPLModule(
        total_num_steps=total_num_steps,
        save_dir=args.save,
        oc_file=args.cfg)
    trainer_kwargs = pl_module.set_trainer_kwargs(
        devices=args.gpus,
        accumulate_grad_batches=accumulate_grad_batches,
    )
    trainer = Trainer(**trainer_kwargs)
    if args.pretrained:
        pretrained_ckpt_name = pytorch_state_dict_name
        if not os.path.exists(os.path.join(pretrained_checkpoints_dir, pretrained_ckpt_name)):
            s3_download_pretrained_ckpt(ckpt_name=pretrained_ckpt_name,
                                        save_dir=pretrained_checkpoints_dir,
                                        exist_ok=False)
        state_dict = torch.load(os.path.join(pretrained_checkpoints_dir, pretrained_ckpt_name),
                                map_location=torch.device("cpu"))
        pl_module.torch_nn_module.load_state_dict(state_dict=state_dict)
        trainer.test(model=pl_module,
                     datamodule=dm)
    elif args.test:
        assert args.ckpt_name is not None, f"args.ckpt_name is required for test!"
        ckpt_path = os.path.join(pl_module.save_dir, "checkpoints", args.ckpt_name)
        trainer.test(model=pl_module,
                     datamodule=dm,
                     ckpt_path=ckpt_path)
    else:
        if args.ckpt_name is not None:
            ckpt_path = os.path.join(pl_module.save_dir, "checkpoints", args.ckpt_name)
            if not os.path.exists(ckpt_path):
                warnings.warn(f"ckpt {ckpt_path} not exists! Start training from epoch 0.")
                ckpt_path = None
        else:
            ckpt_path = None
        trainer.fit(model=pl_module,
                    datamodule=dm,
                    ckpt_path=ckpt_path)
        state_dict = pl_ckpt_to_pytorch_state_dict(checkpoint_path=trainer.checkpoint_callback.best_model_path,
                                                   map_location=torch.device("cpu"),
                                                   delete_prefix_len=len("torch_nn_module."))
        torch.save(state_dict, os.path.join(pl_module.save_dir, "checkpoints", pytorch_state_dict_name))
        trainer.test(ckpt_path="best",
                     datamodule=dm)

if __name__ == "__main__":
    main()



import argparse
from earthformer.datasets.sevir.sevir_torch_wrap import download_SEVIR


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", default="sevir", type=str)
    parser.add_argument("--save", default=None, type=str)
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.dataset == "sevir":
        download_SEVIR(save_dir=args.save)
    else:
        raise ValueError(f"Wrong dataset name {args.dataset}! Must be one of ('sevir', 'sevir_lr').")


if __name__ == "__main__":
    main()



import os
import torch
from pytorch_lightning import LightningDataModule
from torch.nn import functional as F
from torch.utils.data import Dataset, DataLoader
from .nbody_mnist import NBodyMovingMNISTIterator


class NBodyMovingMNISTTorchDataset(Dataset):

    def __init__(self,
                 data_path=None,
                 force_regenerate=False,
                 num_samples=10000,
                 digit_num=None,
                 img_size=64,
                 raw_img_size=128,
                 seq_len=20,
                 raw_seq_len_multiplier=10,
                 distractor_num=None,
                 distractor_size=5,
                 max_velocity_scale=3.6,
                 initial_velocity_range=(0.0, 3.6),
                 random_acceleration_range=(0.0, 0.0),
                 scale_variation_range=(1.0, 1.0),
                 rotation_angle_range=(-0, 0),
                 illumination_factor_range=(1.0, 1.0),
                 period=5,
                 global_rotation_prob=0.5,
                 index_range=(0, 40000),
                 mnist_data_path=None,
                 rescale_01=True,
                 # N-Body params
                 nbody_acc_mode=None,
                 nbody_G=1.0,
                 nbody_softening_distance=1.0,
                 nbody_mass=None,
                 # for debugging
                 return_raw_seq=False,
                 ):
        r"""
        Parameters
        ----------
        data_path: str
            Path of stored data. If not exists, generated to this path.
        force_regenerate: bool
            If True, force to regenerate data even if data stored in data_path already exists.
        num_samples: int
            Number of sequences
        digit_num: int
            Number of digits
        img_size: int
            Sampled sequences have shape (batch_size, seq_len, img_size, img_size, 1)
        raw_img_size: int
            The original image size for generating sequences
            before downsampling (or upsampling) to `img_size`,
            since the raw digits are rather hard to scale.
        seq_len: int
            Sampled sequences have shape (batch_size, seq_len, img_size, img_size, 1)
        raw_seq_len_multiplier: int
            The original generated sequences have higher temporal resolution for
            more precise calculation of acceleration effect. `raw_seq_len == seq_len * raw_seq_len_multiplier'.
        distractor_num: int
            Number of distractors
        distractor_size: int
            Size of the distractors
        max_velocity_scale: float
            Maximum scale of the velocity
        initial_velocity_range: tuple
        random_acceleration_range: tuple
        scale_variation_range: tuple
        rotation_angle_range: tuple
        period: int
            period of the
        index_range: Tuple[int]
            The range of indices of used mnist images
        mnist_data_path: str
            The path storing raw mnist digits images
        rescale_01:
            pixel value range from [0, 255] or [0, 1]
        nbody_acc_mode
            Should be one of [None, "r0", "r1", "r2"]
        nbody_G
            Newton's Gravitational constant
        nbody_softening_distance
            softening length
        nbody_mass
        """
        super(NBodyMovingMNISTTorchDataset, self).__init__()
        # for debugging
        self.return_raw_seq = return_raw_seq
        self.rescale_01 = rescale_01

        self.img_size = img_size
        self.raw_img_size = raw_img_size
        self.seq_len = seq_len
        raw_seq_len = seq_len * raw_seq_len_multiplier
        self.raw_seq_len_multiplier = raw_seq_len_multiplier
        self.raw_seq_len = raw_seq_len
        self.raw_period = period * raw_seq_len_multiplier

        self.num_samples = num_samples

        self.nbody_iterator = NBodyMovingMNISTIterator(
            digit_num=digit_num,
            distractor_num=distractor_num,
            img_size=raw_img_size,
            distractor_size=distractor_size,
            max_velocity_scale=max_velocity_scale,
            initial_velocity_range=initial_velocity_range,
            random_acceleration_range=random_acceleration_range,
            scale_variation_range=scale_variation_range,
            rotation_angle_range=rotation_angle_range,
            illumination_factor_range=illumination_factor_range,
            period=self.raw_period,
            global_rotation_prob=global_rotation_prob,
            index_range=index_range,
            mnist_data_path=mnist_data_path,
            # N-Body params
            nbody_acc_mode=nbody_acc_mode,
            nbody_G=nbody_G,
            nbody_softening_distance=nbody_softening_distance,
            nbody_mass=nbody_mass,)

        if data_path is None:
            data_path = f"generated_nbody.npz"
        self.data_path = data_path
        if force_regenerate or not os.path.exists(data_path):
            print(f"Generating new dataset to {data_path}")
            self.save(file=data_path)
        else:
            print(f"Loading existing dataset from {data_path}")
        num_samples, raw_seq_len = self.load(file=data_path)
        assert num_samples == self.num_samples, f"`num_samples` mismatch! " \
                                                f"loaded num_samples = {num_samples}, " \
                                                f"self.num_samples = {self.num_samples}"
        assert raw_seq_len == self.raw_seq_len, f"`raw_seq_len` mismatch! " \
                                                f"loaded raw_seq_len = {raw_seq_len}, " \
                                                f"self.raw_seq_len = {self.raw_seq_len}"

    def save(self, file=None):
        self.nbody_iterator.save(seqlen=self.raw_seq_len,
                                 num_samples=self.num_samples,
                                 file=file)

    def load(self, file=None):
        num_samples, raw_seq_len = self.nbody_iterator.load(file=file)
        return num_samples, raw_seq_len

    def __getitem__(self, index):
        r"""
        The layout of self.nbody_iterator.sample() is "NTHW"
        The layout of returned seq is "NTHWC" with C = 1
        """
        seq = self.nbody_iterator.sample(batch_size=1,
                                         seqlen=self.raw_seq_len,
                                         replay_index=index)
        ret = torch.from_numpy(seq[0, ...])
        if self.return_raw_seq:
            ret = ret.unsqueeze(-1)
        else:
            ret = ret[::self.raw_seq_len_multiplier, ...]
            ret = F.interpolate(input=ret.unsqueeze(1),
                                size=(self.img_size, self.img_size),
                                mode="bicubic",
                                align_corners=False)
            ret = ret.squeeze(1).unsqueeze(-1)
        ret.clamp_(0.0, 255.0)
        if self.rescale_01:
            ret /= 255.0
        return ret

    def __len__(self):
        return self.nbody_iterator.replay_numsamples


class NBodyMovingMNISTLightningDataModule(LightningDataModule):

    def __init__(self,
                 data_dir=None,
                 force_regenerate=False,
                 num_train_samples=8100,
                 num_val_samples=900,
                 num_test_samples=1000,
                 digit_num=None,
                 img_size=64,
                 raw_img_size=128,
                 seq_len=20,
                 raw_seq_len_multiplier=5,
                 distractor_num=None,
                 distractor_size=5,
                 max_velocity_scale=2.0,
                 initial_velocity_range=(0.0, 2.0),
                 random_acceleration_range=(0.0, 0.0),
                 scale_variation_range=(1.0, 1.0),
                 rotation_angle_range=(-0, 0),
                 illumination_factor_range=(1.0, 1.0),
                 period=5,
                 global_rotation_prob=0.5,
                 index_range=(0, 40000),
                 mnist_data_path=None,
                 rescale_01=True,
                 # N-Body params
                 nbody_acc_mode="r0",
                 nbody_G=0.05,
                 nbody_softening_distance=10.0,
                 nbody_mass=None,
                 # datamodule_only
                 batch_size=1,
                 num_workers=8,
                 ):
        super(NBodyMovingMNISTLightningDataModule, self).__init__()
        if data_dir is None:
            data_dir = "nbody_datamodule"
        self.data_dir = data_dir
        self.force_regenerate = force_regenerate
        self.num_train_samples = num_train_samples
        self.num_val_samples = num_val_samples
        self.num_test_samples = num_test_samples
        self.digit_num = digit_num
        self.seq_len = seq_len
        self.raw_seq_len_multiplier = raw_seq_len_multiplier
        self.img_size = img_size
        self.raw_img_size = raw_img_size
        self.max_velocity_scale = max_velocity_scale
        self.initial_velocity_range = initial_velocity_range
        self.random_acceleration_range = random_acceleration_range
        self.distractor_num = distractor_num
        self.distractor_size = distractor_size
        self.rotation_angle_range = rotation_angle_range
        self.scale_variation_range = scale_variation_range
        self.illumination_factor_range = illumination_factor_range
        self.period = period
        self.global_rotation_prob = global_rotation_prob
        self.index_range = index_range
        self.mnist_data_path = mnist_data_path
        self.rescale_01 = rescale_01
        # N-Body
        self.nbody_acc_mode = nbody_acc_mode
        self.nbody_G = nbody_G
        self.nbody_softening_distance = nbody_softening_distance
        self.nbody_mass = nbody_mass
        # datamodule_only
        self.batch_size = batch_size
        self.num_workers = num_workers

        self.train_data_path = os.path.join(self.data_dir, "nbody_train.npz")
        self.val_data_path = os.path.join(self.data_dir, "nbody_val.npz")
        self.test_data_path = os.path.join(self.data_dir, "nbody_test.npz")

    def prepare_data(self):
        NBodyMovingMNISTTorchDataset(
            data_path=self.train_data_path,
            force_regenerate=self.force_regenerate,
            num_samples=self.num_train_samples,
            digit_num=self.digit_num,
            seq_len=self.seq_len,
            raw_seq_len_multiplier=self.raw_seq_len_multiplier,
            img_size=self.img_size,
            raw_img_size=self.raw_img_size,
            max_velocity_scale=self.max_velocity_scale,
            initial_velocity_range=self.initial_velocity_range,
            random_acceleration_range=self.random_acceleration_range,
            distractor_num=self.distractor_num,
            distractor_size=self.distractor_size,
            rotation_angle_range=self.rotation_angle_range,
            scale_variation_range=self.scale_variation_range,
            illumination_factor_range=self.illumination_factor_range,
            period=self.period,
            global_rotation_prob=self.global_rotation_prob,
            index_range=self.index_range,
            mnist_data_path=self.mnist_data_path,
            rescale_01=self.rescale_01,
            # N-Body
            nbody_acc_mode=self.nbody_acc_mode,
            nbody_G=self.nbody_G,
            nbody_softening_distance=self.nbody_softening_distance,
            nbody_mass=self.nbody_mass,)
        NBodyMovingMNISTTorchDataset(
            data_path=self.val_data_path,
            force_regenerate=self.force_regenerate,
            num_samples=self.num_val_samples,
            digit_num=self.digit_num,
            seq_len=self.seq_len,
            raw_seq_len_multiplier=self.raw_seq_len_multiplier,
            img_size=self.img_size,
            raw_img_size=self.raw_img_size,
            max_velocity_scale=self.max_velocity_scale,
            initial_velocity_range=self.initial_velocity_range,
            random_acceleration_range=self.random_acceleration_range,
            distractor_num=self.distractor_num,
            distractor_size=self.distractor_size,
            rotation_angle_range=self.rotation_angle_range,
            scale_variation_range=self.scale_variation_range,
            illumination_factor_range=self.illumination_factor_range,
            period=self.period,
            global_rotation_prob=self.global_rotation_prob,
            index_range=self.index_range,
            mnist_data_path=self.mnist_data_path,
            rescale_01=self.rescale_01,
            # N-Body
            nbody_acc_mode=self.nbody_acc_mode,
            nbody_G=self.nbody_G,
            nbody_softening_distance=self.nbody_softening_distance,
            nbody_mass=self.nbody_mass,)
        NBodyMovingMNISTTorchDataset(
            data_path=self.test_data_path,
            force_regenerate=self.force_regenerate,
            num_samples=self.num_test_samples,
            digit_num=self.digit_num,
            seq_len=self.seq_len,
            raw_seq_len_multiplier=self.raw_seq_len_multiplier,
            img_size=self.img_size,
            raw_img_size=self.raw_img_size,
            max_velocity_scale=self.max_velocity_scale,
            initial_velocity_range=self.initial_velocity_range,
            random_acceleration_range=self.random_acceleration_range,
            distractor_num=self.distractor_num,
            distractor_size=self.distractor_size,
            rotation_angle_range=self.rotation_angle_range,
            scale_variation_range=self.scale_variation_range,
            illumination_factor_range=self.illumination_factor_range,
            period=self.period,
            global_rotation_prob=self.global_rotation_prob,
            index_range=self.index_range,
            mnist_data_path=self.mnist_data_path,
            rescale_01=self.rescale_01,
            # N-Body
            nbody_acc_mode=self.nbody_acc_mode,
            nbody_G=self.nbody_G,
            nbody_softening_distance=self.nbody_softening_distance,
            nbody_mass=self.nbody_mass,)

    def setup(self, stage = None):
        if stage in (None, "fit"):
            self.nbody_train = NBodyMovingMNISTTorchDataset(
                data_path=self.train_data_path,
                force_regenerate=False,
                num_samples=self.num_train_samples,
                digit_num=self.digit_num,
                seq_len=self.seq_len,
                raw_seq_len_multiplier=self.raw_seq_len_multiplier,
                img_size=self.img_size,
                raw_img_size=self.raw_img_size,
                max_velocity_scale=self.max_velocity_scale,
                initial_velocity_range=self.initial_velocity_range,
                random_acceleration_range=self.random_acceleration_range,
                distractor_num=self.distractor_num,
                distractor_size=self.distractor_size,
                rotation_angle_range=self.rotation_angle_range,
                scale_variation_range=self.scale_variation_range,
                illumination_factor_range=self.illumination_factor_range,
                period=self.period,
                global_rotation_prob=self.global_rotation_prob,
                index_range=self.index_range,
                mnist_data_path=self.mnist_data_path,
                rescale_01=self.rescale_01,
                # N-Body
                nbody_acc_mode=self.nbody_acc_mode,
                nbody_G=self.nbody_G,
                nbody_softening_distance=self.nbody_softening_distance,
                nbody_mass=self.nbody_mass,)
            self.nbody_val = NBodyMovingMNISTTorchDataset(
                data_path=self.val_data_path,
                force_regenerate=False,
                num_samples=self.num_val_samples,
                digit_num=self.digit_num,
                seq_len=self.seq_len,
                raw_seq_len_multiplier=self.raw_seq_len_multiplier,
                img_size=self.img_size,
                raw_img_size=self.raw_img_size,
                max_velocity_scale=self.max_velocity_scale,
                initial_velocity_range=self.initial_velocity_range,
                random_acceleration_range=self.random_acceleration_range,
                distractor_num=self.distractor_num,
                distractor_size=self.distractor_size,
                rotation_angle_range=self.rotation_angle_range,
                scale_variation_range=self.scale_variation_range,
                illumination_factor_range=self.illumination_factor_range,
                period=self.period,
                global_rotation_prob=self.global_rotation_prob,
                index_range=self.index_range,
                mnist_data_path=self.mnist_data_path,
                rescale_01=self.rescale_01,
                # N-Body
                nbody_acc_mode=self.nbody_acc_mode,
                nbody_G=self.nbody_G,
                nbody_softening_distance=self.nbody_softening_distance,
                nbody_mass=self.nbody_mass,)
        if stage in (None, "test"):
            self.nbody_test = NBodyMovingMNISTTorchDataset(
                data_path=self.test_data_path,
                force_regenerate=False,
                num_samples=self.num_test_samples,
                digit_num=self.digit_num,
                seq_len=self.seq_len,
                raw_seq_len_multiplier=self.raw_seq_len_multiplier,
                img_size=self.img_size,
                raw_img_size=self.raw_img_size,
                max_velocity_scale=self.max_velocity_scale,
                initial_velocity_range=self.initial_velocity_range,
                random_acceleration_range=self.random_acceleration_range,
                distractor_num=self.distractor_num,
                distractor_size=self.distractor_size,
                rotation_angle_range=self.rotation_angle_range,
                scale_variation_range=self.scale_variation_range,
                illumination_factor_range=self.illumination_factor_range,
                period=self.period,
                global_rotation_prob=self.global_rotation_prob,
                index_range=self.index_range,
                mnist_data_path=self.mnist_data_path,
                rescale_01=self.rescale_01,
                # N-Body
                nbody_acc_mode=self.nbody_acc_mode,
                nbody_G=self.nbody_G,
                nbody_softening_distance=self.nbody_softening_distance,
                nbody_mass=self.nbody_mass,)
        if stage in (None, "predict"):
            self.nbody_predict = NBodyMovingMNISTTorchDataset(
                data_path=self.test_data_path,
                force_regenerate=False,
                num_samples=self.num_test_samples,
                digit_num=self.digit_num,
                seq_len=self.seq_len,
                raw_seq_len_multiplier=self.raw_seq_len_multiplier,
                img_size=self.img_size,
                raw_img_size=self.raw_img_size,
                max_velocity_scale=self.max_velocity_scale,
                initial_velocity_range=self.initial_velocity_range,
                random_acceleration_range=self.random_acceleration_range,
                distractor_num=self.distractor_num,
                distractor_size=self.distractor_size,
                rotation_angle_range=self.rotation_angle_range,
                scale_variation_range=self.scale_variation_range,
                illumination_factor_range=self.illumination_factor_range,
                period=self.period,
                global_rotation_prob=self.global_rotation_prob,
                index_range=self.index_range,
                mnist_data_path=self.mnist_data_path,
                rescale_01=self.rescale_01,
                # N-Body
                nbody_acc_mode=self.nbody_acc_mode,
                nbody_G=self.nbody_G,
                nbody_softening_distance=self.nbody_softening_distance,
                nbody_mass=self.nbody_mass,)

    def train_dataloader(self):
        return DataLoader(self.nbody_train,
                          batch_size=self.batch_size,
                          shuffle=True,
                          num_workers=self.num_workers)

    def val_dataloader(self):
        return DataLoader(self.nbody_val,
                          batch_size=self.batch_size,
                          shuffle=False,
                          num_workers=self.num_workers)

    def test_dataloader(self):
        return DataLoader(self.nbody_test,
                          batch_size=self.batch_size,
                          shuffle=False,
                          num_workers=self.num_workers)

    def predict_dataloader(self):
        return DataLoader(self.nbody_predict,
                          batch_size=self.batch_size,
                          shuffle=False,
                          num_workers=self.num_workers)



from packaging import version
import torch


if version.parse(torch.__version__) >= version.parse('1.11.0'):
    # Starting from torch>=1.11.0, the attribute `optimizer` is set in SequentialLR. See https://github.com/pytorch/pytorch/pull/67406
    from torch.optim.lr_scheduler import SequentialLR
else:
    from torch.optim.lr_scheduler import _LRScheduler
    from bisect import bisect_right

    class SequentialLR(_LRScheduler):
        """Receives the list of schedulers that is expected to be called sequentially during
        optimization process and milestone points that provides exact intervals to reflect
        which scheduler is supposed to be called at a given epoch.

        Args:
            schedulers (list): List of chained schedulers.
            milestones (list): List of integers that reflects milestone points.

        Example:
            >>> # Assuming optimizer uses lr = 1. for all groups
            >>> # lr = 0.1     if epoch == 0
            >>> # lr = 0.1     if epoch == 1
            >>> # lr = 0.9     if epoch == 2
            >>> # lr = 0.81    if epoch == 3
            >>> # lr = 0.729   if epoch == 4
            >>> scheduler1 = ConstantLR(self.opt, factor=0.1, total_iters=2)
            >>> scheduler2 = ExponentialLR(self.opt, gamma=0.9)
            >>> scheduler = SequentialLR(self.opt, schedulers=[scheduler1, scheduler2], milestones=[2])
            >>> for epoch in range(100):
            >>>     train(...)
            >>>     validate(...)
            >>>     scheduler.step()
        """

        def __init__(self, optimizer, schedulers, milestones, last_epoch=-1, verbose=False):
            for scheduler_idx in range(1, len(schedulers)):
                if (schedulers[scheduler_idx].optimizer != schedulers[0].optimizer):
                    raise ValueError(
                        "Sequential Schedulers expects all schedulers to belong to the same optimizer, but "
                        "got schedulers at index {} and {} to be different".format(0, scheduler_idx)
                    )
            if (len(milestones) != len(schedulers) - 1):
                raise ValueError(
                    "Sequential Schedulers expects number of schedulers provided to be one more "
                    "than the number of milestone points, but got number of schedulers {} and the "
                    "number of milestones to be equal to {}".format(len(schedulers), len(milestones))
                )
            self.optimizer = optimizer
            self._schedulers = schedulers
            self._milestones = milestones
            self.last_epoch = last_epoch + 1

        def step(self):
            self.last_epoch += 1
            idx = bisect_right(self._milestones, self.last_epoch)
            if idx > 0 and self._milestones[idx - 1] == self.last_epoch:
                self._schedulers[idx].step(0)
            else:
                self._schedulers[idx].step()

        def state_dict(self):
            """Returns the state of the scheduler as a :class:`dict`.

            It contains an entry for every variable in self.__dict__ which
            is not the optimizer.
            The wrapped scheduler states will also be saved.
            """
            state_dict = {key: value for key, value in self.__dict__.items() if key not in ('optimizer', '_schedulers')}
            state_dict['_schedulers'] = [None] * len(self._schedulers)

            for idx, s in enumerate(self._schedulers):
                state_dict['_schedulers'][idx] = s.state_dict()

            return state_dict

        def load_state_dict(self, state_dict):
            """Loads the schedulers state.

            Args:
                state_dict (dict): scheduler state. Should be an object returned
                    from a call to :meth:`state_dict`.
            """
            _schedulers = state_dict.pop('_schedulers')
            self.__dict__.update(state_dict)
            # Restore state_dict keys in order to prevent side effects
            # https://github.com/pytorch/pytorch/issues/32756
            state_dict['_schedulers'] = _schedulers

            for idx, s in enumerate(_schedulers):
                self._schedulers[idx].load_state_dict(s)

def warmup_lambda(warmup_steps, min_lr_ratio=0.1):
    def ret_lambda(epoch):
        if epoch <= warmup_steps:
            return min_lr_ratio + (1.0 - min_lr_ratio) * epoch / warmup_steps
        else:
            return 1.0
    return ret_lambda



from typing import  Optional, Sequence
import numpy as np
import torch
from torchmetrics import Metric
from .skill_scores import _threshold
from ..datasets.sevir.sevir_dataloader import SEVIRDataLoader


class SEVIRSkillScore(Metric):
    r"""
    The calculation of skill scores in SEVIR challenge is slightly different:
        `mCSI = sum(mCSI_t) / T`
    See https://github.com/MIT-AI-Accelerator/sevir_challenges/blob/dev/radar_nowcasting/RadarNowcastBenchmarks.ipynb for more details.
    """
    full_state_update: bool = True

    def __init__(self,
                 layout: str = "NHWT",
                 mode: str = "0",
                 seq_len: Optional[int] = None,
                 preprocess_type: str = "sevir",
                 threshold_list: Sequence[int] = (16, 74, 133, 160, 181, 219),
                 metrics_list: Sequence[str] = ("csi", "bias", "sucr", "pod"),
                 eps: float = 1e-4,
                 dist_sync_on_step: bool = False,
                 ):
        r"""
        Parameters
        ----------
        seq_len
        layout
        mode:   str
            Should be in ("0", "1", "2")
            "0":
                cumulates hits/misses/fas of all test pixels
                score_avg takes average over all thresholds
                return
                    score_thresh shape = (1, )
                    score_avg shape = (1, )
            "1":
                cumulates hits/misses/fas of each step
                score_avg takes average over all thresholds while keeps the seq_len dim
                return
                    score_thresh shape = (seq_len, )
                    score_avg shape = (seq_len, )
            "2":
                cumulates hits/misses/fas of each step
                score_avg takes average over all thresholds, then takes average over the seq_len dim
                return
                    score_thresh shape = (1, )
                    score_avg shape = (1, )
        preprocess_type
        threshold_list
        dist_sync_on_step
        """
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.layout = layout
        self.preprocess_type = preprocess_type
        self.threshold_list = threshold_list
        self.metrics_list = metrics_list
        self.eps = eps
        self.mode = mode
        self.seq_len = seq_len
        if mode in ("0", ):
            self.keep_seq_len_dim = False
            state_shape = (len(self.threshold_list), )
        elif mode in ("1", "2"):
            self.keep_seq_len_dim = True
            assert isinstance(self.seq_len, int), "seq_len must be provided when we need to keep seq_len dim."
            state_shape = (len(self.threshold_list), self.seq_len)

        else:
            raise NotImplementedError(f"mode {mode} not supported!")

        self.add_state("hits",
                       default=torch.zeros(state_shape),
                       dist_reduce_fx="sum")
        self.add_state("misses",
                       default=torch.zeros(state_shape),
                       dist_reduce_fx="sum")
        self.add_state("fas",
                       default=torch.zeros(state_shape),
                       dist_reduce_fx="sum")

    @property
    def hits_misses_fas_reduce_dims(self):
        if not hasattr(self, "_hits_misses_fas_reduce_dims"):
            seq_dim = self.layout.find('T')
            self._hits_misses_fas_reduce_dims = list(range(len(self.layout)))
            if self.keep_seq_len_dim:
                self._hits_misses_fas_reduce_dims.pop(seq_dim)
        return self._hits_misses_fas_reduce_dims

    @staticmethod
    def pod(hits, misses, fas, eps):
        return hits / (hits + misses + eps)

    @staticmethod
    def sucr(hits, misses, fas, eps):
        return hits / (hits + fas + eps)

    @staticmethod
    def csi(hits, misses, fas, eps):
        return hits / (hits + misses + fas + eps)

    @staticmethod
    def bias(hits, misses, fas, eps):
        bias = (hits + fas) / (hits + misses + eps)
        logbias = torch.pow(bias / torch.log(torch.tensor(2.0)), 2.0)
        return logbias

    def calc_seq_hits_misses_fas(self, pred, target, threshold):
        """
        Parameters
        ----------
        pred, target:   torch.Tensor
        threshold:  int

        Returns
        -------
        hits, misses, fas:  torch.Tensor
            each has shape (seq_len, )
        """
        with torch.no_grad():
            t, p = _threshold(target, pred, threshold)
            hits = torch.sum(t * p, dim=self.hits_misses_fas_reduce_dims).int()
            misses = torch.sum(t * (1 - p), dim=self.hits_misses_fas_reduce_dims).int()
            fas = torch.sum((1 - t) * p, dim=self.hits_misses_fas_reduce_dims).int()
        return hits, misses, fas

    def preprocess(self, pred, target):
        if self.preprocess_type == "sevir":
            pred = SEVIRDataLoader.process_data_dict_back(
                data_dict={'vil': pred.detach().float()})['vil']
            target = SEVIRDataLoader.process_data_dict_back(
                data_dict={'vil': target.detach().float()})['vil']
        else:
            raise NotImplementedError
        return pred, target

    def update(self, pred: torch.Tensor, target: torch.Tensor):
        pred, target = self.preprocess(pred, target)
        for i, threshold in enumerate(self.threshold_list):
            hits, misses, fas = self.calc_seq_hits_misses_fas(pred, target, threshold)
            self.hits[i] += hits
            self.misses[i] += misses
            self.fas[i] += fas

    def compute(self):
        metrics_dict = {
            'pod': self.pod,
            'csi': self.csi,
            'sucr': self.sucr,
            'bias': self.bias}
        ret = {}
        for threshold in self.threshold_list:
            ret[threshold] = {}
        ret["avg"] = {}
        for metrics in self.metrics_list:
            if self.keep_seq_len_dim:
                score_avg = np.zeros((self.seq_len, ))
            else:
                score_avg = 0
            # shape = (len(threshold_list), seq_len) if self.keep_seq_len_dim,
            # else shape = (len(threshold_list),)
            scores = metrics_dict[metrics](self.hits, self.misses, self.fas, self.eps)
            scores = scores.detach().cpu().numpy()
            for i, threshold in enumerate(self.threshold_list):
                if self.keep_seq_len_dim:
                    score = scores[i]  # shape = (seq_len, )
                else:
                    score = scores[i].item()  # shape = (1, )
                if self.mode in ("0", "1"):
                    ret[threshold][metrics] = score
                elif self.mode in ("2", ):
                    ret[threshold][metrics] = np.mean(score).item()
                else:
                    raise NotImplementedError
                score_avg += score
            score_avg /= len(self.threshold_list)
            if self.mode in ("0", "1"):
                ret["avg"][metrics] = score_avg
            elif self.mode in ("2",):
                ret["avg"][metrics] = np.mean(score_avg).item()
            else:
                raise NotImplementedError
        return ret



import warnings
from shutil import copyfile
import inspect
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.optim.lr_scheduler import LambdaLR, CosineAnnealingLR
import torchmetrics
import pytorch_lightning as pl
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor, DeviceStatsMonitor, Callback
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from omegaconf import OmegaConf
import os
import argparse
from einops import rearrange
from pytorch_lightning import Trainer, seed_everything
from earthformer.config import cfg
from earthformer.utils.optim import SequentialLR, warmup_lambda
from earthformer.utils.utils import get_parameter_names
from earthformer.utils.layout import layout_to_in_out_slice
from earthformer.visualization.nbody import save_example_vis_results
from earthformer.cuboid_transformer.cuboid_transformer import CuboidTransformerModel
from earthformer.datasets.moving_mnist.moving_mnist import MovingMNISTDataModule
from earthformer.utils.apex_ddp import ApexDDPStrategy


_curr_dir = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))
exps_dir = os.path.join(_curr_dir, "experiments")

class CuboidMovingMNISTPLModule(pl.LightningModule):

    def __init__(self,
                 total_num_steps: int,
                 oc_file: str = None,
                 save_dir: str = None):
        super(CuboidMovingMNISTPLModule, self).__init__()
        self._max_train_iter = total_num_steps
        if oc_file is not None:
            oc_from_file = OmegaConf.load(open(oc_file, "r"))
        else:
            oc_from_file = None
        oc = self.get_base_config(oc_from_file=oc_from_file)
        model_cfg = OmegaConf.to_object(oc.model)
        num_blocks = len(model_cfg["enc_depth"])
        if isinstance(model_cfg["self_pattern"], str):
            enc_attn_patterns = [model_cfg["self_pattern"]] * num_blocks
        else:
            enc_attn_patterns = OmegaConf.to_container(model_cfg["self_pattern"])
        if isinstance(model_cfg["cross_self_pattern"], str):
            dec_self_attn_patterns = [model_cfg["cross_self_pattern"]] * num_blocks
        else:
            dec_self_attn_patterns = OmegaConf.to_container(model_cfg["cross_self_pattern"])
        if isinstance(model_cfg["cross_pattern"], str):
            dec_cross_attn_patterns = [model_cfg["cross_pattern"]] * num_blocks
        else:
            dec_cross_attn_patterns = OmegaConf.to_container(model_cfg["cross_pattern"])

        self.torch_nn_module = CuboidTransformerModel(
            input_shape=model_cfg["input_shape"],
            target_shape=model_cfg["target_shape"],
            base_units=model_cfg["base_units"],
            block_units=model_cfg["block_units"],
            scale_alpha=model_cfg["scale_alpha"],
            enc_depth=model_cfg["enc_depth"],
            dec_depth=model_cfg["dec_depth"],
            enc_use_inter_ffn=model_cfg["enc_use_inter_ffn"],
            dec_use_inter_ffn=model_cfg["dec_use_inter_ffn"],
            dec_hierarchical_pos_embed=model_cfg["dec_hierarchical_pos_embed"],
            downsample=model_cfg["downsample"],
            downsample_type=model_cfg["downsample_type"],
            enc_attn_patterns=enc_attn_patterns,
            dec_self_attn_patterns=dec_self_attn_patterns,
            dec_cross_attn_patterns=dec_cross_attn_patterns,
            dec_cross_last_n_frames=model_cfg["dec_cross_last_n_frames"],
            dec_use_first_self_attn=model_cfg["dec_use_first_self_attn"],
            num_heads=model_cfg["num_heads"],
            attn_drop=model_cfg["attn_drop"],
            proj_drop=model_cfg["proj_drop"],
            ffn_drop=model_cfg["ffn_drop"],
            upsample_type=model_cfg["upsample_type"],
            ffn_activation=model_cfg["ffn_activation"],
            gated_ffn=model_cfg["gated_ffn"],
            norm_layer=model_cfg["norm_layer"],
            # global vectors
            num_global_vectors=model_cfg["num_global_vectors"],
            use_dec_self_global=model_cfg["use_dec_self_global"],
            dec_self_update_global=model_cfg["dec_self_update_global"],
            use_dec_cross_global=model_cfg["use_dec_cross_global"],
            use_global_vector_ffn=model_cfg["use_global_vector_ffn"],
            use_global_self_attn=model_cfg["use_global_self_attn"],
            separate_global_qkv=model_cfg["separate_global_qkv"],
            global_dim_ratio=model_cfg["global_dim_ratio"],
            # initial_downsample
            initial_downsample_type=model_cfg["initial_downsample_type"],
            initial_downsample_activation=model_cfg["initial_downsample_activation"],
            # initial_downsample_type=="conv"
            initial_downsample_scale=model_cfg["initial_downsample_scale"],
            initial_downsample_conv_layers=model_cfg["initial_downsample_conv_layers"],
            final_upsample_conv_layers=model_cfg["final_upsample_conv_layers"],
            # misc
            padding_type=model_cfg["padding_type"],
            z_init_method=model_cfg["z_init_method"],
            checkpoint_level=model_cfg["checkpoint_level"],
            pos_embed_type=model_cfg["pos_embed_type"],
            use_relative_pos=model_cfg["use_relative_pos"],
            self_attn_use_final_proj=model_cfg["self_attn_use_final_proj"],
            # initialization
            attn_linear_init_mode=model_cfg["attn_linear_init_mode"],
            ffn_linear_init_mode=model_cfg["ffn_linear_init_mode"],
            conv_init_mode=model_cfg["conv_init_mode"],
            down_up_linear_init_mode=model_cfg["down_up_linear_init_mode"],
            norm_init_mode=model_cfg["norm_init_mode"],
        )

        self.total_num_steps = total_num_steps
        if oc_file is not None:
            oc_from_file = OmegaConf.load(open(oc_file, "r"))
        else:
            oc_from_file = None
        oc = self.get_base_config(oc_from_file=oc_from_file)
        self.save_hyperparameters(oc)
        self.oc = oc
        # layout
        self.in_len = oc.layout.in_len
        self.out_len = oc.layout.out_len
        self.layout = oc.layout.layout
        # optimization
        self.max_epochs = oc.optim.max_epochs
        self.optim_method = oc.optim.method
        self.lr = oc.optim.lr
        self.wd = oc.optim.wd
        # lr_scheduler
        self.total_num_steps = total_num_steps
        self.lr_scheduler_mode = oc.optim.lr_scheduler_mode
        self.warmup_percentage = oc.optim.warmup_percentage
        self.min_lr_ratio = oc.optim.min_lr_ratio
        # logging
        self.save_dir = save_dir
        self.logging_prefix = oc.logging.logging_prefix
        # visualization
        self.train_example_data_idx_list = list(oc.vis.train_example_data_idx_list)
        self.val_example_data_idx_list = list(oc.vis.val_example_data_idx_list)
        self.test_example_data_idx_list = list(oc.vis.test_example_data_idx_list)
        self.eval_example_only = oc.vis.eval_example_only

        self.configure_save(cfg_file_path=oc_file)

        self.valid_mse = torchmetrics.MeanSquaredError()
        self.valid_mae = torchmetrics.MeanAbsoluteError()
        self.valid_ssim = torchmetrics.StructuralSimilarityIndexMeasure()
        self.test_mse = torchmetrics.MeanSquaredError()
        self.test_mae = torchmetrics.MeanAbsoluteError()
        self.test_ssim = torchmetrics.StructuralSimilarityIndexMeasure()

    def configure_save(self, cfg_file_path=None):
        self.save_dir = os.path.join(exps_dir, self.save_dir)
        os.makedirs(self.save_dir, exist_ok=True)
        self.scores_dir = os.path.join(self.save_dir, 'scores')
        os.makedirs(self.scores_dir, exist_ok=True)
        if cfg_file_path is not None:
            cfg_file_target_path = os.path.join(self.save_dir, "cfg.yaml")
            if (not os.path.exists(cfg_file_target_path)) or \
                    (not os.path.samefile(cfg_file_path, cfg_file_target_path)):
                copyfile(cfg_file_path, cfg_file_target_path)
        self.example_save_dir = os.path.join(self.save_dir, "examples")
        os.makedirs(self.example_save_dir, exist_ok=True)

    def get_base_config(self, oc_from_file=None):
        oc = OmegaConf.create()
        oc.layout = self.get_layout_config()
        oc.optim = self.get_optim_config()
        oc.logging = self.get_logging_config()
        oc.trainer = self.get_trainer_config()
        oc.vis = self.get_vis_config()
        oc.model = self.get_model_config()
        if oc_from_file is not None:
            # oc = apply_omegaconf_overrides(oc, oc_from_file)
            oc = OmegaConf.merge(oc, oc_from_file)
        return oc

    @classmethod
    def get_model_config(cls):
        cfg = OmegaConf.create()
        height = 64
        width = 64
        in_len = 10
        out_len = 10
        data_channels = 1
        cfg.input_shape = (in_len, height, width, data_channels)
        cfg.target_shape = (out_len, height, width, data_channels)

        cfg.base_units = 64
        cfg.block_units = None # multiply by 2 when downsampling in each layer
        cfg.scale_alpha = 1.0

        cfg.enc_depth = [1, 1]
        cfg.dec_depth = [1, 1]
        cfg.enc_use_inter_ffn = True
        cfg.dec_use_inter_ffn = True
        cfg.dec_hierarchical_pos_embed = True

        cfg.downsample = 2
        cfg.downsample_type = "patch_merge"
        cfg.upsample_type = "upsample"

        cfg.num_global_vectors = 8
        cfg.use_dec_self_global = True
        cfg.dec_self_update_global = True
        cfg.use_dec_cross_global = True
        cfg.use_global_vector_ffn = True
        cfg.use_global_self_attn = False
        cfg.separate_global_qkv = False
        cfg.global_dim_ratio = 1

        cfg.self_pattern = 'axial'
        cfg.cross_self_pattern = 'axial'
        cfg.cross_pattern = 'cross_1x1'
        cfg.dec_cross_last_n_frames = None

        cfg.attn_drop = 0.1
        cfg.proj_drop = 0.1
        cfg.ffn_drop = 0.1
        cfg.num_heads = 4

        cfg.ffn_activation = 'gelu'
        cfg.gated_ffn = False
        cfg.norm_layer = 'layer_norm'
        cfg.padding_type = 'zeros'
        cfg.pos_embed_type = "t+hw"
        cfg.use_relative_pos = True
        cfg.self_attn_use_final_proj = True
        cfg.dec_use_first_self_attn = False

        cfg.z_init_method = 'zeros'  # The method for initializing the first input of the decoder
        cfg.initial_downsample_type = "conv"
        cfg.initial_downsample_activation = "leaky"
        cfg.initial_downsample_scale = 2
        cfg.initial_downsample_conv_layers = 2
        cfg.final_upsample_conv_layers = 1
        cfg.checkpoint_level = 2
        # initialization
        cfg.attn_linear_init_mode = "0"
        cfg.ffn_linear_init_mode = "0"
        cfg.conv_init_mode = "0"
        cfg.down_up_linear_init_mode = "0"
        cfg.norm_init_mode = "0"
        return cfg

    @staticmethod
    def get_layout_config():
        oc = OmegaConf.create()
        oc.in_len = 10
        oc.out_len = 10
        oc.layout = "NTHWC"  # The layout of the data, not the model
        return oc

    @staticmethod
    def get_optim_config():
        oc = OmegaConf.create()
        oc.seed = None
        oc.total_batch_size = 32
        oc.micro_batch_size = 8

        oc.method = "adamw"
        oc.lr = 1E-3
        oc.wd = 1E-5
        oc.gradient_clip_val = 1.0
        oc.max_epochs = 50
        # scheduler
        oc.warmup_percentage = 0.2
        oc.lr_scheduler_mode = "cosine"  # Can be strings like 'linear', 'cosine', 'platue'
        oc.min_lr_ratio = 0.1
        oc.warmup_min_lr_ratio = 0.1
        # early stopping
        oc.early_stop = False
        oc.early_stop_mode = "min"
        oc.early_stop_patience = 5
        oc.save_top_k = 1
        return oc

    @staticmethod
    def get_logging_config():
        oc = OmegaConf.create()
        oc.logging_prefix = "MovingMNIST"
        oc.monitor_lr = True
        oc.monitor_device = False
        oc.track_grad_norm = -1
        cfg.use_wandb = False
        return oc

    @staticmethod
    def get_trainer_config():
        oc = OmegaConf.create()
        oc.check_val_every_n_epoch = 1
        oc.log_step_ratio = 0.001  # Logging every 1% of the total training steps per epoch
        oc.precision = 32
        return oc

    @staticmethod
    def get_vis_config():
        oc = OmegaConf.create()
        oc.train_example_data_idx_list = [0, ]
        oc.val_example_data_idx_list = [0, ]
        oc.test_example_data_idx_list = [0, ]
        oc.eval_example_only = False
        return oc

    def configure_optimizers(self):
        # Configure the optimizer. Disable the weight decay for layer norm weights and all bias terms.
        decay_parameters = get_parameter_names(self.torch_nn_module, [nn.LayerNorm])
        decay_parameters = [name for name in decay_parameters if "bias" not in name]
        optimizer_grouped_parameters = [{
            'params': [p for n, p in self.torch_nn_module.named_parameters() if n in decay_parameters],
            'weight_decay': self.oc.optim.wd
        }, {
            'params': [p for n, p in self.torch_nn_module.named_parameters() if n not in decay_parameters],
            'weight_decay': 0.0
        }]

        if self.oc.optim.method == 'adamw':
            optimizer = torch.optim.AdamW(params=optimizer_grouped_parameters,
                                          lr=self.oc.optim.lr,
                                          weight_decay=self.oc.optim.wd)
        else:
            raise NotImplementedError

        warmup_iter = int(np.round(self.oc.optim.warmup_percentage * self.total_num_steps))

        if self.oc.optim.lr_scheduler_mode == 'cosine':
            warmup_scheduler = LambdaLR(optimizer,
                                        lr_lambda=warmup_lambda(warmup_steps=warmup_iter,
                                                                min_lr_ratio=self.oc.optim.warmup_min_lr_ratio))
            cosine_scheduler = CosineAnnealingLR(optimizer,
                                                 T_max=(self.total_num_steps - warmup_iter),
                                                 eta_min=self.oc.optim.min_lr_ratio * self.oc.optim.lr)
            lr_scheduler = SequentialLR(optimizer, schedulers=[warmup_scheduler, cosine_scheduler],
                                        milestones=[warmup_iter])
            lr_scheduler_config = {
                'scheduler': lr_scheduler,
                'interval': 'step',
                'frequency': 1,
            }
        else:
            raise NotImplementedError
        return {'optimizer': optimizer, 'lr_scheduler': lr_scheduler_config}

    def set_trainer_kwargs(self, **kwargs):
        r"""
        Default kwargs used when initializing pl.Trainer
        """
        checkpoint_callback = ModelCheckpoint(
            monitor="valid_loss_epoch",
            dirpath=os.path.join(self.save_dir, "checkpoints"),
            filename="model-{epoch:03d}",
            save_top_k=self.oc.optim.save_top_k,
            save_last=True,
            mode="min",
        )
        callbacks = kwargs.pop("callbacks", [])
        assert isinstance(callbacks, list)
        for ele in callbacks:
            assert isinstance(ele, Callback)
        callbacks += [checkpoint_callback, ]
        if self.oc.logging.monitor_lr:
            callbacks += [LearningRateMonitor(logging_interval='step'), ]
        if self.oc.logging.monitor_device:
            callbacks += [DeviceStatsMonitor(), ]
        if self.oc.optim.early_stop:
            callbacks += [EarlyStopping(monitor="valid_loss_epoch",
                                        min_delta=0.0,
                                        patience=self.oc.optim.early_stop_patience,
                                        verbose=False,
                                        mode=self.oc.optim.early_stop_mode), ]

        logger = kwargs.pop("logger", [])
        tb_logger = pl_loggers.TensorBoardLogger(save_dir=self.save_dir)
        csv_logger = pl_loggers.CSVLogger(save_dir=self.save_dir)
        logger += [tb_logger, csv_logger]
        if self.oc.logging.use_wandb:
            wandb_logger = pl_loggers.WandbLogger(project=self.oc.logging.logging_prefix,
                                                  save_dir=self.save_dir)
            logger += [wandb_logger, ]

        log_every_n_steps = max(1, int(self.oc.trainer.log_step_ratio * self.total_num_steps))
        trainer_init_keys = inspect.signature(Trainer).parameters.keys()
        ret = dict(
            callbacks=callbacks,
            # log
            logger=logger,
            log_every_n_steps=log_every_n_steps,
            track_grad_norm=self.oc.logging.track_grad_norm,
            # save
            default_root_dir=self.save_dir,
            # ddp
            accelerator="gpu",
            # strategy="ddp",
            strategy=ApexDDPStrategy(find_unused_parameters=False, delay_allreduce=True),
            # optimization
            max_epochs=self.oc.optim.max_epochs,
            check_val_every_n_epoch=self.oc.trainer.check_val_every_n_epoch,
            gradient_clip_val=self.oc.optim.gradient_clip_val,
            # NVIDIA amp
            precision=self.oc.trainer.precision,
        )
        oc_trainer_kwargs = OmegaConf.to_object(self.oc.trainer)
        oc_trainer_kwargs = {key: val for key, val in oc_trainer_kwargs.items() if key in trainer_init_keys}
        ret.update(oc_trainer_kwargs)
        ret.update(kwargs)
        return ret

    @classmethod
    def get_total_num_steps(
            cls,
            num_samples: int,
            total_batch_size: int,
            epoch: int = None):
        r"""
        Parameters
        ----------
        num_samples:    int
            The number of samples of the datasets. `num_samples / micro_batch_size` is the number of steps per epoch.
        total_batch_size:   int
            `total_batch_size == micro_batch_size * world_size * grad_accum`
        """
        if epoch is None:
            epoch = cls.get_optim_config().max_epochs
        return int(epoch * num_samples / total_batch_size)

    @staticmethod
    def get_movingmnist_datamodule(micro_batch_size: int = 1):
        dm = MovingMNISTDataModule(batch_size=micro_batch_size)
        return dm

    def forward(self, batch):
        in_seq, target_seq = batch
        pred_seq = self.torch_nn_module(in_seq)
        loss = F.mse_loss(pred_seq, target_seq)
        return pred_seq, loss

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat, loss = self(batch)
        self.save_vis_step_end(
            batch_idx=batch_idx,
            in_seq=x, target_seq=y,
            pred_seq=y_hat,
            mode="train"
        )
        self.log('train_loss', loss, on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        B, T_out, H, W, C = y.shape
        if not self.eval_example_only or batch_idx in self.val_example_data_idx_list:
            y_hat, _ = self(batch)
            self.save_vis_step_end(
                batch_idx=batch_idx,
                in_seq=x, target_seq=y,
                pred_seq=y_hat,
                mode="val"
            )
            if self.precision == 16:
                y_hat = y_hat.float()
            step_mse = self.valid_mse(y_hat, y) * H * W
            step_mae = self.valid_mae(y_hat, y) * H * W
            y_hat = rearrange(y_hat,
                              "b t h w c -> (b t) c h w")
            y = rearrange(y,
                          "b t h w c -> (b t) c h w")
            step_ssim = self.valid_ssim(y_hat, y)

            self.log('valid_mse_step', step_mse, prog_bar=True, on_step=True, on_epoch=False)
            self.log('valid_mae_step', step_mae, prog_bar=True, on_step=True, on_epoch=False)
            self.log('valid_ssim_step', step_ssim, prog_bar=True, on_step=True, on_epoch=False)
        return H, W

    def validation_epoch_end(self, outputs):
        H, W = outputs[0]
        frame_mse = self.valid_mse.compute() * H * W
        frame_mae = self.valid_mae.compute() * H * W
        valid_loss = frame_mse
        epoch_ssim = self.valid_ssim.compute()

        self.log('valid_loss_epoch', valid_loss, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_mse_epoch', frame_mse, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_mae_epoch', frame_mae, prog_bar=True, on_step=False, on_epoch=True)
        self.log('valid_ssim_epoch', epoch_ssim, prog_bar=True, on_step=False, on_epoch=True)
        self.valid_mse.reset()
        self.valid_mae.reset()
        self.valid_ssim.reset()

    def test_step(self, batch, batch_idx):
        x, y = batch
        B, T_out, H, W, C = y.shape
        if not self.eval_example_only or batch_idx in self.test_example_data_idx_list:
            y_hat, _ = self(batch)
            self.save_vis_step_end(
                batch_idx=batch_idx,
                in_seq=x, target_seq=y,
                pred_seq=y_hat,
                mode="test"
            )
            if self.precision == 16:
                y_hat = y_hat.float()
            step_mse = self.test_mse(y_hat, y) * H * W
            step_mae = self.test_mae(y_hat, y) * H * W
            y_hat = rearrange(y_hat,
                              "b t h w c -> (b t) c h w")
            y = rearrange(y,
                          "b t h w c -> (b t) c h w")
            step_ssim = self.test_ssim(y_hat, y)

            self.log('test_mse_step', step_mse, prog_bar=True, on_step=True, on_epoch=False)
            self.log('test_mae_step', step_mae, prog_bar=True, on_step=True, on_epoch=False)
            self.log('test_ssim_step', step_ssim, prog_bar=True, on_step=True, on_epoch=False)
        return H, W

    def test_epoch_end(self, outputs):
        H, W = outputs[0]
        frame_mse = self.test_mse.compute() * H * W
        frame_mae = self.test_mae.compute() * H * W
        epoch_ssim = self.test_ssim.compute()

        self.log('test_mse_epoch', frame_mse, prog_bar=True, on_step=False, on_epoch=True)
        self.log('test_mae_epoch', frame_mae, prog_bar=True, on_step=False, on_epoch=True)
        self.log('test_ssim_epoch', epoch_ssim, prog_bar=True, on_step=False, on_epoch=True)
        self.test_mse.reset()
        self.test_mae.reset()
        self.test_ssim.reset()

    def save_vis_step_end(
            self,
            batch_idx: int,
            in_seq: torch.Tensor, target_seq: torch.Tensor,
            pred_seq: torch.Tensor,
            mode: str = "train"):

        if self.local_rank == 0:
            if mode == "train":
                example_data_idx_list = self.train_example_data_idx_list
            elif mode == "val":
                example_data_idx_list = self.val_example_data_idx_list
            elif mode == "test":
                example_data_idx_list = self.test_example_data_idx_list
            else:
                raise ValueError(f"Wrong mode {mode}! Must be in ['train', 'val', 'test'].")
            if batch_idx in example_data_idx_list:
                micro_batch_size = in_seq.shape[self.layout.find("N")]
                data_idx = int(batch_idx * micro_batch_size)
                save_example_vis_results(
                    save_dir=self.example_save_dir,
                    save_prefix=f'{mode}_epoch_{self.current_epoch}_data_{data_idx}',
                    in_seq=in_seq.detach().float().cpu().numpy(),
                    target_seq=target_seq.detach().float().cpu().numpy(),
                    pred_seq=pred_seq.detach().float().cpu().numpy(),
                    layout=self.layout,
                    plot_stride=1,
                    label=self.oc.logging.logging_prefix)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--save', default='tmp_mnist', type=str)
    parser.add_argument('--gpus', default=1, type=int)
    parser.add_argument('--cfg', default=None, type=str)
    parser.add_argument('--test', action='store_true')
    parser.add_argument('--ckpt_name', default=None, type=str,
                        help='The model checkpoint trained on N-body MovingMNIST.')
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.cfg is not None:
        oc_from_file = OmegaConf.load(open(args.cfg, "r"))
        total_batch_size = oc_from_file.optim.total_batch_size
        micro_batch_size = oc_from_file.optim.micro_batch_size
        max_epochs = oc_from_file.optim.max_epochs
        seed = oc_from_file.optim.seed
    else:
        micro_batch_size = 1
        total_batch_size = int(micro_batch_size * args.gpus)
        max_epochs = None
        seed = 0
    seed_everything(seed, workers=True)
    dm = CuboidMovingMNISTPLModule.get_movingmnist_datamodule(micro_batch_size=micro_batch_size)
    dm.prepare_data()
    dm.setup()
    accumulate_grad_batches = total_batch_size // (micro_batch_size * args.gpus)
    total_num_steps = CuboidMovingMNISTPLModule.get_total_num_steps(
        epoch=max_epochs,
        num_samples=dm.num_train_samples,
        total_batch_size=total_batch_size,
    )
    pl_module = CuboidMovingMNISTPLModule(
        total_num_steps=total_num_steps,
        save_dir=args.save,
        oc_file=args.cfg)
    trainer_kwargs = pl_module.set_trainer_kwargs(
        devices=args.gpus,
        accumulate_grad_batches=accumulate_grad_batches,
    )
    trainer = Trainer(**trainer_kwargs)
    if args.test:
        assert args.ckpt_name is not None, f"args.ckpt_name is required for test!"
        ckpt_path = os.path.join(pl_module.save_dir, "checkpoints", args.ckpt_name)
        trainer.test(model=pl_module,
                     datamodule=dm,
                     ckpt_path=ckpt_path)
    else:
        if args.ckpt_name is not None:
            ckpt_path = os.path.join(pl_module.save_dir, "checkpoints", args.ckpt_name)
            if not os.path.exists(ckpt_path):
                warnings.warn(f"ckpt {ckpt_path} not exists! Start training from epoch 0.")
                ckpt_path = None
        else:
            ckpt_path = None
        trainer.fit(model=pl_module,
                    datamodule=dm,
                    ckpt_path=ckpt_path)
        trainer.test(ckpt_path="best",
                     datamodule=dm)

if __name__ == "__main__":
    main()



import datetime

import numpy as np
from matplotlib import pyplot as plt

from earthformer.datasets.sevir.sevir_dataloader import SEVIR_CATALOG, SEVIR_DATA_DIR, SEVIR_RAW_SEQ_LEN, \
    SEVIR_LR_CATALOG, SEVIR_LR_DATA_DIR, SEVIR_LR_RAW_SEQ_LEN, SEVIRDataLoader


def report_SEVIR_statistics(
        dataset="sevir",
        sanity_check=True,
        hist_save_path="tmp_sevir_hist.png"):
    r"""
    Report important statistics of SEVIR dataset, including:
    - The distribution of pixel values (from 0 to 255).

    Refer to https://discuss.pytorch.org/t/plot-a-histogram-for-multiple-images-full-dataset/67600
    """
    if sanity_check:
        start_date = datetime.datetime(2019, 5, 27)
        train_val_split_date = datetime.datetime(2019, 5, 29)
        train_test_split_date = datetime.datetime(2019, 6, 1)
        end_date = datetime.datetime(2019, 6, 3)
    else:
        # total number of event = 14926 for the whole training set
        # val percentage is 20% in original SEVIR paper
        # this split results in 11906 / 14926 = 79.77%
        # if use (2019, 2, 1), the ratio becomes 12489 / 14926 = 83.67%
        start_date = None
        train_val_split_date = datetime.datetime(2019, 1, 1)
        # train_val_split_date = datetime.datetime(2019, 2, 1)
        train_test_split_date = datetime.datetime(2019, 6, 1)
        end_date = None
    if dataset == "sevir":
        catalog_path = SEVIR_CATALOG
        data_dir = SEVIR_DATA_DIR
        raw_seq_len = SEVIR_RAW_SEQ_LEN
    elif dataset == "sevir_lr":
        catalog_path = SEVIR_LR_CATALOG
        data_dir = SEVIR_LR_DATA_DIR
        raw_seq_len = SEVIR_LR_RAW_SEQ_LEN
    else:
        raise ValueError(f"Invalid dataset: {dataset}")

    batch_size = 1
    data_types = ["vil", ]
    layout = "NHWT"
    seq_len = raw_seq_len
    stride = seq_len
    sample_mode = "sequent"

    train_dataloader = SEVIRDataLoader(
        sevir_catalog=catalog_path, sevir_data_dir=data_dir,
        raw_seq_len=raw_seq_len, split_mode="uneven",
        data_types=data_types, seq_len=seq_len, stride=stride,
        sample_mode=sample_mode, batch_size=batch_size, layout=layout,
        num_shard=1, rank=0,
        start_date=start_date, end_date=train_val_split_date)
    # val_dataloader = SEVIRDataLoader(
    #     sevir_catalog=catalog_path, sevir_data_dir=data_dir,
    #     raw_seq_len=raw_seq_len, split_mode="uneven",
    #     data_types=data_types, seq_len=seq_len, stride=stride,
    #     sample_mode=sample_mode, batch_size=batch_size, layout=layout,
    #     num_shard=1, rank=0,
    #     start_date=train_val_split_date, end_date=train_test_split_date)
    # test_dataloader = SEVIRDataLoader(
    #     sevir_catalog=catalog_path, sevir_data_dir=data_dir,
    #     raw_seq_len=raw_seq_len, split_mode="uneven",
    #     data_types=data_types, seq_len=seq_len, stride=stride,
    #     sample_mode=sample_mode, batch_size=batch_size, layout=layout,
    #     num_shard=1, rank=0,
    #     start_date=train_test_split_date, end_date=end_date)

    data_loader = train_dataloader
    # data_loader = val_dataloader
    # data_loader = test_dataloader

    num_bins = 256
    count = np.zeros(num_bins)
    for data_idx, data in enumerate(data_loader):
        data_seq = data['vil']
        hist = np.histogram(data_seq, bins=num_bins, range=[0, 255])
        count += hist[0]
    bins = hist[1]
    fig = plt.figure()
    plt.bar(bins[:-1], count, color='b', alpha=0.5)
    plt.savefig(hist_save_path)


import numpy as np
from scipy.stats import hmean
import torch
from torchmetrics import Metric
from einops import rearrange
from .earthnet_toolkit.parallel_score import CubeCalculator as EN_CubeCalculator
from ...metrics.torchmetrics_wo_compute import MetricsUpdateWithoutCompute


class EarthNet2021Score(Metric):

    default_layout = "NHWCT"
    default_channel_axis = 3
    channels = 4

    def __init__(self,
                 layout: str = "NTHWC",
                 eps: float = 1e-4,
                 dist_sync_on_step: bool = False, ):
        super(EarthNet2021Score, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.layout = layout
        self.eps = eps

        self.add_state("MAD",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("OLS",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("EMD",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("SSIM",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        # does not count if NaN
        self.add_state("num_MAD",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_OLS",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_EMD",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_SSIM",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")

    @property
    def einops_default_layout(self):
        if not hasattr(self, "_einops_default_layout"):
            self._einops_default_layout = " ".join(self.default_layout)
        return self._einops_default_layout

    @property
    def einops_layout(self):
        if not hasattr(self, "_einops_layout"):
            self._einops_layout = " ".join(self.layout)
        return self._einops_layout

    def update(self, pred: torch.Tensor, target: torch.Tensor, mask: torch.Tensor = None):
        r"""

        Parameters
        ----------
        pred, target:   torch.Tensor
            With the first dim as batch dim, and 4 channels (RGB and infrared)
        mask:   torch.Tensor
            With the first dim as batch dim, and 1 channel
        """
        pred_np = rearrange(pred.detach(), f"{self.einops_layout} -> {self.einops_default_layout}").cpu().numpy()
        target_np = rearrange(target.detach(), f"{self.einops_layout} -> {self.einops_default_layout}").cpu().numpy()
        # layout = "NHWCT"
        if mask is None:
            mask_np = np.ones_like(target_np)
        else:
            mask_np = torch.repeat_interleave(
                rearrange(1 - mask.detach(), f"{self.einops_layout} -> {self.einops_default_layout}"),
                repeats=self.channels, dim=self.default_channel_axis).cpu().numpy()
        for preds, targs, masks in zip(pred_np, target_np, mask_np):
            # Code is adapted from `load_file()` in ./earthnet_toolkit/parallel_score.py
            preds[preds < 0] = 0
            preds[preds > 1] = 1

            targs[np.isnan(targs)] = 0
            targs[targs > 1] = 1
            targs[targs < 0] = 0

            ndvi_preds = ((preds[:, :, 3, :] - preds[:, :, 2, :]) / (preds[:, :, 3, :] + preds[:, :, 2, :] + 1e-6))[:,
                         :, np.newaxis, :]
            ndvi_targs = ((targs[:, :, 3, :] - targs[:, :, 2, :]) / (targs[:, :, 3, :] + targs[:, :, 2, :] + 1e-6))[:,
                         :, np.newaxis, :]
            ndvi_masks = masks[:, :, 0, :][:, :, np.newaxis, :]
            # Code is adapted from `get_scores()` in ./earthnet_toolkit/parallel_score.py
            debug_info = {}
            mad, debug_info["MAD"] = EN_CubeCalculator.MAD(preds, targs, masks)
            ols, debug_info["OLS"] = EN_CubeCalculator.OLS(ndvi_preds, ndvi_targs, ndvi_masks)
            emd, debug_info["EMD"] = EN_CubeCalculator.EMD(ndvi_preds, ndvi_targs, ndvi_masks)
            ssim, debug_info["SSIM"] = EN_CubeCalculator.SSIM(preds, targs, masks)
            # does not count if NaN
            if mad is not None and not np.isnan(mad):
                self.MAD += mad
                self.num_MAD += 1
            if ols is not None and not np.isnan(ols):
                self.OLS += ols
                self.num_OLS += 1
            if emd is not None and not np.isnan(emd):
                self.EMD += emd
                self.num_EMD += 1
            if ssim is not None and not np.isnan(ssim):
                self.SSIM += ssim
                self.num_SSIM += 1

    def compute(self):
        MAD_mean = (self.MAD / (self.num_MAD + self.eps)).cpu().item()
        OLS_mean = (self.OLS / (self.num_OLS + self.eps)).cpu().item()
        EMD_mean = (self.EMD / (self.num_EMD + self.eps)).cpu().item()
        SSIM_mean = (self.SSIM / (self.num_SSIM + self.eps)).cpu().item()
        ENS = hmean([MAD_mean, OLS_mean, EMD_mean, SSIM_mean])
        return {
            "MAD": MAD_mean,
            "OLS": OLS_mean,
            "EMD": EMD_mean,
            "SSIM":SSIM_mean,
            "EarthNetScore": ENS,
        }

class EarthNet2021ScoreUpdateWithoutCompute(MetricsUpdateWithoutCompute):

    default_layout = "NHWCT"
    default_channel_axis = 3
    channels = 4

    def __init__(self,
                 layout: str = "NTHWC",
                 eps: float = 1e-4,
                 dist_sync_on_step: bool = False, ):
        super(EarthNet2021ScoreUpdateWithoutCompute, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.layout = layout
        self.eps = eps

        self.add_state("MAD",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("OLS",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("EMD",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        self.add_state("SSIM",
                       default=torch.tensor(0.0),
                       dist_reduce_fx="sum")
        # does not count if NaN
        self.add_state("num_MAD",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_OLS",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_EMD",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")
        self.add_state("num_SSIM",
                       default=torch.tensor(0),
                       dist_reduce_fx="sum")

    @property
    def einops_default_layout(self):
        if not hasattr(self, "_einops_default_layout"):
            self._einops_default_layout = " ".join(self.default_layout)
        return self._einops_default_layout

    @property
    def einops_layout(self):
        if not hasattr(self, "_einops_layout"):
            self._einops_layout = " ".join(self.layout)
        return self._einops_layout

    def update(self, pred: torch.Tensor, target: torch.Tensor, mask: torch.Tensor = None):
        r"""

        Parameters
        ----------
        pred, target:   torch.Tensor
            With the first dim as batch dim, and 4 channels (RGB and infrared)
        mask:   torch.Tensor
            With the first dim as batch dim, and 1 channel
        """
        pred_np = rearrange(pred.detach(), f"{self.einops_layout} -> {self.einops_default_layout}").cpu().numpy()
        target_np = rearrange(target.detach(), f"{self.einops_layout} -> {self.einops_default_layout}").cpu().numpy()
        # layout = "NHWCT"
        if mask is None:
            mask_np = np.ones_like(target_np)
        else:
            mask_np = torch.repeat_interleave(
                rearrange(1 - mask.detach(), f"{self.einops_layout} -> {self.einops_default_layout}"),
                repeats=self.channels, dim=self.default_channel_axis).cpu().numpy()
        for preds, targs, masks in zip(pred_np, target_np, mask_np):
            # Code is adapted from `load_file()` in ./earthnet_toolkit/parallel_score.py
            preds[preds < 0] = 0
            preds[preds > 1] = 1

            targs[np.isnan(targs)] = 0
            targs[targs > 1] = 1
            targs[targs < 0] = 0

            ndvi_preds = ((preds[:, :, 3, :] - preds[:, :, 2, :]) / (preds[:, :, 3, :] + preds[:, :, 2, :] + 1e-6))[:,
                         :, np.newaxis, :]
            ndvi_targs = ((targs[:, :, 3, :] - targs[:, :, 2, :]) / (targs[:, :, 3, :] + targs[:, :, 2, :] + 1e-6))[:,
                         :, np.newaxis, :]
            ndvi_masks = masks[:, :, 0, :][:, :, np.newaxis, :]
            # Code is adapted from `get_scores()` in ./earthnet_toolkit/parallel_score.py
            debug_info = {}
            mad, debug_info["MAD"] = EN_CubeCalculator.MAD(preds, targs, masks)
            ols, debug_info["OLS"] = EN_CubeCalculator.OLS(ndvi_preds, ndvi_targs, ndvi_masks)
            emd, debug_info["EMD"] = EN_CubeCalculator.EMD(ndvi_preds, ndvi_targs, ndvi_masks)
            ssim, debug_info["SSIM"] = EN_CubeCalculator.SSIM(preds, targs, masks)
            # does not count if NaN
            if mad is not None and not np.isnan(mad):
                self.MAD += mad
                self.num_MAD += 1
            if ols is not None and not np.isnan(ols):
                self.OLS += ols
                self.num_OLS += 1
            if emd is not None and not np.isnan(emd):
                self.EMD += emd
                self.num_EMD += 1
            if ssim is not None and not np.isnan(ssim):
                self.SSIM += ssim
                self.num_SSIM += 1

    def compute(self):
        MAD_mean = (self.MAD / (self.num_MAD + self.eps)).cpu().item()
        OLS_mean = (self.OLS / (self.num_OLS + self.eps)).cpu().item()
        EMD_mean = (self.EMD / (self.num_EMD + self.eps)).cpu().item()
        SSIM_mean = (self.SSIM / (self.num_SSIM + self.eps)).cpu().item()
        ENS = hmean([MAD_mean, OLS_mean, EMD_mean, SSIM_mean])
        return {
            "MAD": MAD_mean,
            "OLS": OLS_mean,
            "EMD": EMD_mean,
            "SSIM":SSIM_mean,
            "EarthNetScore": ENS,
        }



from typing import Tuple, Optional, Union
import numpy as np
import torch
from torchmetrics import Metric
from ..datasets.enso.enso_dataloader import NINO_WINDOW_T, scale_back_sst


def compute_enso_score(
        y_pred, y_true,
        acc_weight: Optional[Union[str, np.ndarray, torch.Tensor]] = None):
    r"""

    Parameters
    ----------
    y_pred: torch.Tensor
    y_true: torch.Tensor
    acc_weight: Optional[Union[str, np.ndarray, torch.Tensor]]
        None:   not used
        default:    use default acc_weight specified at https://tianchi.aliyun.com/competition/entrance/531871/information
        np.ndarray: custom weights

    Returns
    -------
    acc
    rmse
    """
    pred = y_pred - y_pred.mean(dim=0, keepdim=True)  # (N, 24)
    true = y_true - y_true.mean(dim=0, keepdim=True)  # (N, 24)
    cor = (pred * true).sum(dim=0) / (torch.sqrt(torch.sum(pred**2, dim=0) * torch.sum(true**2, dim=0)) + 1e-6)

    if acc_weight is None:
        acc = cor.sum()
    else:
        nino_out_len = y_true.shape[-1]
        if acc_weight == "default":
            acc_weight = torch.tensor([1.5] * 4 + [2] * 7 + [3] * 7 + [4] * (nino_out_len - 18))[:nino_out_len] \
                         * torch.log(torch.arange(nino_out_len) + 1)
        elif isinstance(acc_weight, np.ndarray):
            acc_weight = torch.from_numpy(acc_weight[:nino_out_len])
        elif isinstance(acc_weight, torch.Tensor):
            acc_weight = acc_weight[:nino_out_len]
        else:
            raise ValueError(f"Invalid acc_weight {acc_weight}!")
        acc_weight = acc_weight.to(y_pred)
        acc = (acc_weight * cor).sum()
    rmse = torch.mean((y_pred - y_true)**2, dim=0).sqrt().sum()
    return acc, rmse

def sst_to_nino(sst: torch.Tensor,
                normalize_sst: bool = True,
                detach: bool = True):
    r"""

    Parameters
    ----------
    sst:    torch.Tensor
        Shape = (N, T, H, W)

    Returns
    -------
    nino_index: torch.Tensor
        Shape = (N, T-NINO_WINDOW_T+1)
    """
    if detach:
        nino_index = sst.detach()
    else:
        nino_index = sst
    if normalize_sst:
        nino_index = scale_back_sst(nino_index)
    nino_index = nino_index[:, :, 10:13, 19:30].mean(dim=[2, 3])  # (N, 26)
    nino_index = nino_index.unfold(dimension=1, size=NINO_WINDOW_T, step=1).mean(dim=2)  # (N, 24)
    return nino_index

class ENSOScore(Metric):

    def __init__(self,
                 layout="NTHW",
                 out_len=26,
                 normalize_sst=True):
        super(ENSOScore, self).__init__()
        assert layout in ["NTHW", "NTHWC"], f"layout {layout} not supported"
        self.layout = layout
        self.normalize_sst = normalize_sst
        self.out_len = out_len
        self.nino_out_len = out_len - NINO_WINDOW_T + 1
        self.nino_weight = torch.from_numpy(np.array([1.5]*4 + [2]*7 + [3]*7 + [4]*(self.nino_out_len-18))
                                            * np.log(np.arange(self.nino_out_len)+1))

        self.add_state("sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("num_pixels", default=torch.tensor(0), dist_reduce_fx="sum")

        self.add_state("nino_preds", default=[], dist_reduce_fx="cat")
        self.add_state("nino_target", default=[], dist_reduce_fx="cat")
        # self.add_state("nino_preds", default=torch.zeros(0, 24), dist_reduce_fx="cat")
        # self.add_state("nino_target", default=torch.zeros(0, 24), dist_reduce_fx="cat")

    def update(self,
               preds: torch.Tensor, target: torch.Tensor,
               nino_preds:torch.Tensor = None, nino_target: torch.Tensor = None) -> None:
        r"""
        Parameters
        ----------
        preds
            Shape = (N, T, H, W) if self.layout == "NTHW"
                or  (N, T, H, W, 1) if self.layout == "NTHWC"
        target
            Shape = (N, T, H, W) if self.layout == "NTHW"
                or  (N, T, H, W, 1) if self.layout == "NTHWC"
        nino_preds
            Shape = (N, T-NINO_WINDOW_T+1)
        nino_target
            Shape = (N, T-NINO_WINDOW_T+1)
        Returns
        -------
        mse
        """
        if self.layout == "NTHWC":
            preds = preds[..., 0]
            target = target[..., 0]
        diff = preds - target
        sum_squared_error = torch.sum(diff * diff)
        num_pixels = target.numel()
        self.sum_squared_error += sum_squared_error
        self.num_pixels += num_pixels

        if nino_preds is None:
            nino_preds = sst_to_nino(sst=preds,
                                     normalize_sst=self.normalize_sst)
        if nino_target is None:
            nino_target = sst_to_nino(sst=target,
                                      normalize_sst=self.normalize_sst)
        nino_preds_list = [ele for ele in nino_preds]
        nino_target_list = [ele for ele in nino_target]
        self.nino_preds.extend(nino_preds_list)
        self.nino_target.extend(nino_target_list)
        # self.nino_preds = torch.cat((self.nino_preds, nino_preds))
        # self.nino_target = torch.cat((self.nino_target, nino_target))

    def compute(self) -> Tuple[float]:
        mse = self.sum_squared_error / self.num_pixels
        # print(f"self.nino_preds.shape = {self.nino_preds.shape}")
        y_pred=torch.stack(self.nino_preds, dim=0)
        y_true=torch.stack(self.nino_target, dim=0)
        # y_pred=self.nino_preds
        # y_true=self.nino_target
        acc, nino_rmse = compute_enso_score(
            y_pred=y_pred, y_true=y_true,
            acc_weight=self.nino_weight)
        return acc.cpu().item(), mse.cpu().item(),
